############################################################################
# Original work Copyright 2017 Palantir Technologies, Inc.                 #
# Original work licensed under the MIT License.                            #
# See ThirdPartyNotices.txt in the project root for license information.   #
# All modifications Copyright (c) Open Law Library. All rights reserved.   #
#                                                                          #
# Licensed under the Apache License, Version 2.0 (the "License")           #
# you may not use this file except in compliance with the License.         #
# You may obtain a copy of the License at                                  #
#                                                                          #
#     http: // www.apache.org/licenses/LICENSE-2.0                         #
#                                                                          #
# Unless required by applicable law or agreed to in writing, software      #
# distributed under the License is distributed on an "AS IS" BASIS,        #
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. #
# See the License for the specific language governing permissions and      #
# limitations under the License.                                           #
############################################################################
"""
A collection of URI utilities with logic built on the VSCode URI library.

https://github.com/Microsoft/vscode-uri/blob/e59cab84f5df6265aed18ae5f43552d3eef13bb9/lib/index.ts
"""
from typing import Optional, Tuple

import re
from urllib import parse

from pygls import IS_WIN

RE_DRIVE_LETTER_PATH = re.compile(r"^\/[a-zA-Z]:")

URLParts = Tuple[str, str, str, str, str, str]


def _normalize_win_path(path: str):
    netloc = ""

    # normalize to fwd-slashes on windows,
    # on other systems bwd-slashes are valid
    # filename character, eg /f\oo/ba\r.txt
    if IS_WIN:
        path = path.replace("\\", "/")

    # check for authority as used in UNC shares
    # or use the path as given
    if path[:2] == "//":
        idx = path.index("/", 2)
        if idx == -1:
            netloc = path[2:]
        else:
            netloc = path[2:idx]
            path = path[idx:]

    # Ensure that path starts with a slash
    # or that it is at least a slash
    if not path.startswith("/"):
        path = "/" + path

    # Normalize drive paths to lower case
    if RE_DRIVE_LETTER_PATH.match(path):
        path = path[0] + path[1].lower() + path[2:]

    return path, netloc


def from_fs_path(path: str):
    """Returns a URI for the given filesystem path."""
    try:
        scheme = "file"
        params, query, fragment = "", "", ""
        path, netloc = _normalize_win_path(path)
        return urlunparse((scheme, netloc, path, params, query, fragment))
    except (AttributeError, TypeError):
        return None


def to_fs_path(uri: str):
    """
    Returns the filesystem path of the given URI.

    Will handle UNC paths and normalize windows drive letters to lower-case.
    Also uses the platform specific path separator. Will *not* validate the
    path for invalid characters and semantics.
    Will *not* look at the scheme of this URI.
    """
    try:
        # scheme://netloc/path;parameters?query#fragment
        scheme, netloc, path, _, _, _ = urlparse(uri)

        if netloc and path and scheme == "file":
            # unc path: file://shares/c$/far/boo
            value = f"//{netloc}{path}"

        elif RE_DRIVE_LETTER_PATH.match(path):
            # windows drive letter: file:///C:/far/boo
            value = path[1].lower() + path[2:]

        else:
            # Other path
            value = path

        if IS_WIN:
            value = value.replace("/", "\\")

        return value
    except TypeError:
        return None


def uri_scheme(uri: str):
    try:
        return urlparse(uri)[0]
    except (TypeError, IndexError):
        return None


# TODO: Use `URLParts` type
def uri_with(
    uri: str,
    scheme: Optional[str] = None,
    netloc: Optional[str] = None,
    path: Optional[str] = None,
    params: Optional[str] = None,
    query: Optional[str] = None,
    fragment: Optional[str] = None,
):
    """
    Return a URI with the given part(s) replaced.
    Parts are decoded / encoded.
    """
    old_scheme, old_netloc, old_path, old_params, old_query, old_fragment = urlparse(
        uri
    )

    if path is None:
        raise Exception("`path` must not be None")

    path, _ = _normalize_win_path(path)
    return urlunparse(
        (
            scheme or old_scheme,
            netloc or old_netloc,
            path or old_path,
            params or old_params,
            query or old_query,
            fragment or old_fragment,
        )
    )


def urlparse(uri: str):
    """Parse and decode the parts of a URI."""
    scheme, netloc, path, params, query, fragment = parse.urlparse(uri)
    return (
        parse.unquote(scheme),
        parse.unquote(netloc),
        parse.unquote(path),
        parse.unquote(params),
        parse.unquote(query),
        parse.unquote(fragment),
    )


def urlunparse(parts: URLParts) -> str:
    """Unparse and encode parts of a URI."""
    scheme, netloc, path, params, query, fragment = parts

    # Avoid encoding the windows drive letter colon
    if RE_DRIVE_LETTER_PATH.match(path):
        quoted_path = path[:3] + parse.quote(path[3:])
    else:
        quoted_path = parse.quote(path)

    return parse.urlunparse(
        (
            parse.quote(scheme),
            parse.quote(netloc),
            quoted_path,
            parse.quote(params),
            parse.quote(query),
            parse.quote(fragment),
        )
    )



from __future__ import annotations

from _typeshed import StrOrBytesPath
from os import PathLike
from os.path import abspath, expanduser, expandvars
from typing import AnyStr, Union
from typing_extensions import assert_type


def test_str_path(str_path: StrOrBytesPath) -> None:
    # These methods are currently overloaded to work around python/mypy#17952 & python/mypy#11880
    # Let's ensure that they'll still work with a StrOrBytesPath if the workaround is removed

    assert_type(abspath(str_path), Union[str, bytes])
    assert_type(expanduser(str_path), Union[str, bytes])
    assert_type(expandvars(str_path), Union[str, bytes])


# See python/mypy#17952
class MyPathMissingGeneric(PathLike):  # type: ignore # Explicitly testing w/ missing type argument
    def __init__(self, path: str | bytes) -> None:
        super().__init__()
        self.path = path

    def __fspath__(self) -> str | bytes:
        return self.path


# MyPathMissingGeneric could also be fixed by users by adding the missing generic annotation
class MyPathGeneric(PathLike[AnyStr]):
    def __init__(self, path: AnyStr) -> None:
        super().__init__()
        self.path: AnyStr = path

    def __fspath__(self) -> AnyStr:
        return self.path


class MyPathStr(PathLike[str]):
    def __init__(self, path: str) -> None:
        super().__init__()
        self.path = path

    def __fspath__(self) -> str:
        return self.path


abspath(MyPathMissingGeneric("."))
expanduser(MyPathMissingGeneric("."))
expandvars(MyPathMissingGeneric("."))

abspath(MyPathGeneric("."))
expanduser(MyPathGeneric("."))
expandvars(MyPathGeneric("."))

abspath(MyPathStr("."))
expanduser(MyPathStr("."))
expandvars(MyPathStr("."))



from typing import ClassVar, Final

from docutils import parsers

__docformat__: Final = "reStructuredText"

class Parser(parsers.Parser):
    supported: ClassVar[tuple[str, ...]]
    config_section_dependencies: ClassVar[tuple[str, ...]]



from __future__ import annotations

from _typeshed import SupportsDunderGT, SupportsDunderLT, SupportsGetItem
from collections.abc import Callable
from operator import itemgetter
from typing import Any, TypeVar
from typing_extensions import assert_type

_T = TypeVar("_T")


# This should be equivalent to itemgetter().__call__
def standalone_call(obj: SupportsGetItem[Any, _T]) -> _T: ...


# Expected type of itemgetter(1).__call__
expected_type_itemgetter_call: Callable[[SupportsGetItem[int, _T]], _T]  # pyright: ignore[reportGeneralTypeIssues]

# Expecting itemgetter(1) to be assignable to this
# based on the example below: min({"first": 1, "second": 2}.items(), key=itemgetter(1))
# That example and assigning to this variable are what failed in https://github.com/python/mypy/issues/14032
expected_assignable_to: Callable[[tuple[str, int]], SupportsDunderLT[Any] | SupportsDunderGT[Any]]


# Regression tests for https://github.com/python/mypy/issues/14032
# assert_type(itemgetter("first")({"first": 1, "second": 2}), int) # See comment on itemgetter.__call__
assert_type(min({"first": 1, "second": 2}, key=itemgetter(1)), str)
assert_type(min({"first": 1, "second": 2}.items(), key=itemgetter(1)), tuple[str, int])
assert_type(standalone_call({"first": 1, "second": 2}), int)
assert_type(min({"first": 1, "second": 2}, key=standalone_call), str)
assert_type(min({"first": 1, "second": 2}.items(), key=standalone_call), tuple[str, int])

expected_itemgetter_call_type = itemgetter(1).__call__
expected_itemgetter_call_type = itemgetter(1)
expected_assignable_to = itemgetter(1)

expected_itemgetter_call_type = standalone_call
expected_assignable_to = standalone_call



from typing import Final

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["lukes_partitioning"]

D_EDGE_W: Final = "weight"
D_EDGE_VALUE: Final[float]
D_NODE_W: Final = "weight"
D_NODE_VALUE: Final = 1
PKEY: Final = "partitions"
CLUSTER_EVAL_CACHE_SIZE: Final = 2048

@_dispatchable
def lukes_partitioning(G: Graph[_Node], max_size: int, node_weight=None, edge_weight=None): ...



from _typeshed import Incomplete
from collections.abc import Generator, Iterable
from typing import ClassVar, Final, Literal

from docutils import nodes
from docutils.transforms import Transform

__docformat__: Final = "reStructuredText"

class Decorations(Transform):
    default_priority: ClassVar[int]
    def apply(self) -> None: ...
    def generate_header(self) -> None: ...
    def generate_footer(self) -> list[nodes.paragraph] | None: ...

class ExposeInternals(Transform):
    default_priority: ClassVar[int]
    def not_Text(self, node: object) -> bool: ...  # node passing to isinstance() method
    def apply(self) -> None: ...

class Messages(Transform):
    default_priority: ClassVar[int]
    def apply(self) -> None: ...

class FilterMessages(Transform):
    default_priority: ClassVar[int]
    def apply(self) -> None: ...

class TestMessages(Transform):
    __test__: bool
    default_priority: ClassVar[int]
    def apply(self) -> None: ...

class StripComments(Transform):
    default_priority: ClassVar[int]
    def apply(self) -> None: ...

class StripClassesAndElements(Transform):
    default_priority: ClassVar[int]
    strip_elements: set[Incomplete]
    def apply(self) -> None: ...
    def check_classes(self, node: object) -> bool: ...

class SmartQuotes(Transform):
    default_priority: ClassVar[int]
    nodes_to_skip: ClassVar[tuple[type[nodes.Node | nodes.Special], ...]]
    literal_nodes: ClassVar[tuple[type[nodes.Node | nodes.Body], ...]]
    smartquotes_action: ClassVar[str]
    unsupported_languages: set[str]
    def __init__(self, document: nodes.document, startnode: nodes.Node | None) -> None: ...
    def get_tokens(self, txtnodes: Iterable[nodes.Node]) -> Generator[tuple[Literal["literal", "plain"], str]]: ...
    def apply(self) -> None: ...

class Validate(Transform):
    default_priority: ClassVar[int]
    def apply(self) -> None: ...



from _typeshed import Incomplete

__all__ = ["DOCUMENT", "DOCTYPE", "TEXT", "ELEMENT", "COMMENT", "ENTITY", "UNKNOWN", "TreeWalker", "NonRecursiveTreeWalker"]

DOCUMENT: int
DOCTYPE: int
TEXT: int
ELEMENT: int
COMMENT: int
ENTITY: int
UNKNOWN: str
spaceCharacters: str

class TreeWalker:
    tree: Incomplete
    def __init__(self, tree) -> None: ...
    def __iter__(self): ...
    def error(self, msg): ...
    def emptyTag(self, namespace, name, attrs, hasChildren: bool = False) -> None: ...
    def startTag(self, namespace, name, attrs): ...
    def endTag(self, namespace, name): ...
    def text(self, data) -> None: ...
    def comment(self, data): ...
    def doctype(self, name, publicId=None, systemId=None): ...
    def entity(self, name): ...
    def unknown(self, nodeType): ...

class NonRecursiveTreeWalker(TreeWalker):
    def getNodeDetails(self, node) -> None: ...
    def getFirstChild(self, node) -> None: ...
    def getNextSibling(self, node) -> None: ...
    def getParentNode(self, node) -> None: ...
    def __iter__(self): ...



from re import Pattern
from typing import Final
from typing_extensions import TypeAlias

always_safe: Final[str]
urlencoded: Final[set[str]]
INVALID_HEX_PATTERN: Final[Pattern[str]]

_ExplodedQueryString: TypeAlias = list[tuple[str, str]]

def url_encode(params: _ExplodedQueryString) -> str: ...
def url_decode(query: str) -> _ExplodedQueryString: ...
def add_params_to_qs(query: str, params: _ExplodedQueryString) -> str: ...
def add_params_to_uri(uri: str, params: _ExplodedQueryString, fragment: bool = False): ...
def quote(s: str, safe: bytes = b"/") -> str: ...
def unquote(s: str) -> str: ...
def quote_url(s: str) -> str: ...
def extract_params(raw: dict[str, str] | _ExplodedQueryString) -> _ExplodedQueryString: ...
def is_valid_url(url: str, fragments_allowed: bool = True) -> bool: ...



from _typeshed import SupportsRichComparison
from typing import Final, Literal, TypeVar

_SupportsRichComparisonT = TypeVar("_SupportsRichComparisonT", bound=SupportsRichComparison)

Version: Final[str]
__version__: Final[str]
__date__: Final[str]
__min_python_version__: Final[tuple[int, int]]

def cmp(a: _SupportsRichComparisonT, b: _SupportsRichComparisonT) -> Literal[-1, 0, 1]: ...



class XrayCommandListener:
    record_full_documents: bool
    def __init__(self, record_full_documents: bool) -> None: ...
    def started(self, event) -> None: ...
    def succeeded(self, event) -> None: ...
    def failed(self, event) -> None: ...

def patch(record_full_documents: bool = False) -> None: ...



from authlib.common.errors import AuthlibHTTPError

class OAuth1Error(AuthlibHTTPError):
    def __init__(self, description=None, uri=None, status_code=None) -> None: ...
    def get_headers(self): ...

class InsecureTransportError(OAuth1Error):
    error: str
    description: str
    @classmethod
    def check(cls, uri) -> None: ...

class InvalidRequestError(OAuth1Error):
    error: str

class UnsupportedParameterError(OAuth1Error):
    error: str

class UnsupportedSignatureMethodError(OAuth1Error):
    error: str

class MissingRequiredParameterError(OAuth1Error):
    error: str
    def __init__(self, key) -> None: ...

class DuplicatedOAuthProtocolParameterError(OAuth1Error):
    error: str

class InvalidClientError(OAuth1Error):
    error: str
    status_code: int

class InvalidTokenError(OAuth1Error):
    error: str
    description: str
    status_code: int

class InvalidSignatureError(OAuth1Error):
    error: str
    status_code: int

class InvalidNonceError(OAuth1Error):
    error: str
    status_code: int

class AccessDeniedError(OAuth1Error):
    error: str
    description: str

class MethodNotAllowedError(OAuth1Error):
    error: str
    status_code: int



from _typeshed import Incomplete

from reportlab.graphics.barcode.common import Barcode

class _Code39Base(Barcode):
    barWidth: Incomplete
    lquiet: Incomplete
    rquiet: Incomplete
    quiet: int
    gap: Incomplete
    barHeight: Incomplete
    ratio: float
    checksum: int
    bearers: float
    stop: int
    def __init__(self, value: str = "", **args) -> None: ...
    decomposed: Incomplete
    def decompose(self): ...

class Standard39(_Code39Base):
    valid: int
    validated: Incomplete
    def validate(self): ...
    encoded: Incomplete
    def encode(self): ...

class Extended39(_Code39Base):
    valid: int
    validated: Incomplete
    def validate(self): ...
    encoded: str
    def encode(self): ...



from _typeshed import Incomplete

from braintree.webhook_notification import WebhookNotification

text_type = str

class WebhookNotificationGateway:
    gateway: Incomplete
    config: Incomplete
    def __init__(self, gateway) -> None: ...
    def parse(self, signature: str, payload: str) -> WebhookNotification: ...
    def verify(self, challenge: str) -> str: ...



from _typeshed import ReadableBuffer, SupportsRead, SupportsWrite
from collections.abc import Callable, Container, Generator, Mapping
from typing import Any, overload
from typing_extensions import TypeAlias

class ParsingInterrupted(Exception): ...

# dict as attribute value is exclusive to xmlns: https://github.com/bigpick/xmltodict/commit/22541b4874365cb8d2397f23087a866b3081fd9c
_AttrValue: TypeAlias = str | dict[str, str]
_AttrDict: TypeAlias = dict[str, _AttrValue]

class _DictSAXHandler:
    path: list[tuple[str, _AttrDict | None]]
    stack: list[tuple[_AttrDict | None, list[str]]]
    data: list[str]
    item: _AttrDict | None
    item_depth: int
    xml_attribs: bool
    item_callback: Callable[[list[tuple[str, _AttrDict | None]], str | _AttrDict | None], bool]
    attr_prefix: str
    cdata_key: str
    force_cdata: bool | Container[str] | Callable[[tuple[str, _AttrDict | None], str, str], bool]
    cdata_separator: str
    postprocessor: Callable[[list[tuple[str, _AttrDict | None]], str, _AttrValue], tuple[str, _AttrValue]] | None
    dict_constructor: type
    strip_whitespace: bool
    namespace_separator: str
    namespaces: dict[str, str] | None
    namespace_declarations: dict[str, str]
    force_list: bool | Container[str] | Callable[[tuple[str, _AttrDict | None], str, str], bool] | None
    comment_key: str
    def __init__(
        self,
        item_depth: int = 0,
        item_callback: Callable[[list[tuple[str, _AttrDict | None]], str | _AttrDict | None], bool] = ...,
        xml_attribs: bool = True,
        attr_prefix: str = "@",
        cdata_key: str = "#text",
        force_cdata: bool | Container[str] | Callable[[tuple[str, _AttrDict | None], str, str], bool] = False,
        cdata_separator: str = "",
        postprocessor: Callable[[list[tuple[str, _AttrDict | None]], str, _AttrValue], tuple[str, _AttrValue]] | None = None,
        dict_constructor: type = ...,
        strip_whitespace: bool = True,
        namespace_separator: str = ":",
        namespaces: dict[str, str] | None = None,
        force_list: bool | Container[str] | Callable[[tuple[str, _AttrDict | None], str, str], bool] | None = None,
        comment_key: str = "#comment",
    ) -> None: ...
    def startNamespaceDecl(self, prefix: str, uri: str) -> None: ...
    def startElement(self, full_name: str, attrs: dict[str, str] | list[str]) -> None: ...
    def endElement(self, full_name: str) -> None: ...
    def characters(self, data: str) -> None: ...
    def comments(self, data: str) -> None: ...
    def push_data(self, item: _AttrDict | None, key: str, data: str) -> _AttrDict: ...

def parse(
    xml_input: str | ReadableBuffer | SupportsRead[bytes] | Generator[ReadableBuffer],
    encoding: str | None = None,
    expat: Any = ...,
    process_namespaces: bool = False,
    namespace_separator: str = ":",
    disable_entities: bool = True,
    process_comments: bool = False,
    *,
    item_depth: int = 0,
    item_callback: Callable[[list[tuple[str, _AttrDict | None]], str | _AttrDict | None], bool] = ...,
    xml_attribs: bool = True,
    attr_prefix: str = "@",
    cdata_key: str = "#text",
    force_cdata: bool | Container[str] | Callable[[tuple[str, _AttrDict | None], str, str], bool] = False,
    cdata_separator: str = "",
    postprocessor: Callable[[list[tuple[str, _AttrDict | None]], str, _AttrValue], tuple[str, _AttrValue]] | None = None,
    dict_constructor: type = ...,
    strip_whitespace: bool = True,
    namespaces: dict[str, str] | None = None,
    force_list: bool | Container[str] | Callable[[tuple[str, _AttrDict | None], str, str], bool] | None = None,
    comment_key: str = "#comment",
) -> dict[str, Any]: ...
@overload
def unparse(
    input_dict: Mapping[str, Any],
    output: SupportsWrite[bytes] | SupportsWrite[str],
    encoding: str = "utf-8",
    full_document: bool = True,
    short_empty_elements: bool = False,
    comment_key: str = "#comment",
    *,
    attr_prefix: str = "@",
    cdata_key: str = "#text",
    depth: int = 0,
    # preprocessor is called like (preprocessor(key, value) for key, value in input_dict.items()).
    # It is expected to return its input, or a modification thereof
    preprocessor: Callable[[str, Any], tuple[str, Any]] | None = None,
    pretty: bool = False,
    newl: str = "\n",
    indent: str | int = "\t",
    namespace_separator: str = ":",
    namespaces: Mapping[str, str] | None = None,
    expand_iter: str | None = None,
) -> None: ...
@overload
def unparse(
    input_dict: Mapping[str, Any],
    output: None = None,
    encoding: str = "utf-8",
    full_document: bool = True,
    short_empty_elements: bool = False,
    comment_key: str = "#comment",
    *,
    attr_prefix: str = "@",
    cdata_key: str = "#text",
    depth: int = 0,
    # preprocessor is called like (preprocessor(key, value) for key, value in input_dict.items()).
    # It is expected to return its input, or a modification thereof
    preprocessor: Callable[[str, Any], tuple[str, Any]] | None = None,
    pretty: bool = False,
    newl: str = "\n",
    indent: str | int = "\t",
    namespace_separator: str = ":",
    namespaces: Mapping[str, str] | None = None,
    expand_iter: str | None = None,
) -> str: ...



from _typeshed import Unused
from collections.abc import Iterable

from _win32typing import PyIClassFactory, PyIID

def RegisterClassFactories(
    clsids: Iterable[PyIID], flags: int | None = None, clsctx: int | None = None
) -> list[tuple[PyIClassFactory, int]]: ...
def RevokeClassFactories(infos: Iterable[tuple[Unused, int]]) -> None: ...



from hvac.api.system_backend.system_backend_mixin import SystemBackendMixin
from requests.models import _JSON

class Wrapping(SystemBackendMixin):
    def unwrap(self, token=None): ...
    def wrap(self, payload: _JSON | None = None, ttl: int = 60): ...



from _typeshed import Incomplete

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["voterank"]

@_dispatchable
def voterank(G: Graph[_Node], number_of_nodes: int | None = None) -> list[Incomplete]: ...



from _typeshed import StrPath, SupportsRead
from collections.abc import MutableSequence, Sequence
from typing import Any, Literal

g_token_cache: dict[str, list[tuple[str, str]]]  # undocumented
python3: Literal[True]
string_type = str
unicode_type = str

def unicode(x: str, y: str) -> str: ...
def render(
    template: SupportsRead[str] | str | Sequence[tuple[str, str]] = "",
    data: dict[str, Any] = {},
    partials_path: StrPath | None = ".",
    partials_ext: str = "mustache",
    partials_dict: dict[str, str] = {},
    padding: str = "",
    def_ldel: str | None = "{{",
    def_rdel: str | None = "}}",
    scopes: MutableSequence[int] | None = None,
    warn: bool = False,
) -> str: ...



from collections.abc import Callable
from typing import Final

MBLENGTH: Final[dict[int, int]]

class Charset:
    is_default: bool
    def __init__(self, id: int, name: str, collation: str, is_default: bool = False) -> None: ...
    @property
    def encoding(self) -> str: ...
    @property
    def is_binary(self) -> bool: ...

class Charsets:
    def __init__(self) -> None: ...
    def add(self, c: Charset) -> None: ...
    def by_id(self, id: int) -> Charset: ...
    def by_name(self, name: str) -> Charset: ...

_charsets: Charsets
charset_by_name: Callable[[str], Charset]
charset_by_id: Callable[[int], Charset]



from collections.abc import Callable, Sequence
from typing_extensions import TypeAlias

_Algorithm: TypeAlias = Callable[[int], Sequence[int]]

def method(algorithm: _Algorithm, alphabet: str, size: int) -> str: ...



from __future__ import annotations

from collections import Counter, deque
from collections.abc import Mapping as AbcMapping
from collections.abc import MutableMapping as AbcMutableMapping
from collections.abc import MutableSet as AbcMutableSet
from dataclasses import Field
from enum import Enum
from inspect import Signature
from inspect import signature as inspect_signature
from pathlib import Path
from typing import Any, Callable, Iterable, Optional, Tuple, TypeVar, overload

from attrs import Attribute, resolve_types
from attrs import has as attrs_has

from ._compat import (
    ANIES,
    FrozenSetSubscriptable,
    Mapping,
    MutableMapping,
    MutableSequence,
    NoneType,
    OriginAbstractSet,
    OriginMutableSet,
    Sequence,
    Set,
    TypeAlias,
    fields,
    get_final_base,
    get_newtype_base,
    get_origin,
    get_type_alias_base,
    has,
    has_with_generic,
    is_annotated,
    is_bare,
    is_counter,
    is_deque,
    is_frozenset,
    is_generic,
    is_generic_attrs,
    is_hetero_tuple,
    is_literal,
    is_mapping,
    is_mutable_set,
    is_optional,
    is_protocol,
    is_sequence,
    is_tuple,
    is_type_alias,
    is_typeddict,
    is_union_type,
    signature,
)
from .cols import (
    is_namedtuple,
    iterable_unstructure_factory,
    list_structure_factory,
    namedtuple_structure_factory,
    namedtuple_unstructure_factory,
)
from .disambiguators import create_default_dis_func, is_supported_union
from .dispatch import (
    HookFactory,
    MultiStrategyDispatch,
    StructuredValue,
    StructureHook,
    TargetType,
    UnstructuredValue,
    UnstructureHook,
)
from .errors import (
    IterableValidationError,
    IterableValidationNote,
    StructureHandlerNotFoundError,
)
from .fns import Predicate, identity, raise_error
from .gen import (
    AttributeOverride,
    DictStructureFn,
    HeteroTupleUnstructureFn,
    IterableUnstructureFn,
    MappingStructureFn,
    MappingUnstructureFn,
    make_dict_structure_fn,
    make_dict_unstructure_fn,
    make_hetero_tuple_unstructure_fn,
    make_mapping_structure_fn,
    make_mapping_unstructure_fn,
)
from .gen.typeddicts import make_dict_structure_fn as make_typeddict_dict_struct_fn
from .gen.typeddicts import make_dict_unstructure_fn as make_typeddict_dict_unstruct_fn

__all__ = ["UnstructureStrategy", "BaseConverter", "Converter", "GenConverter"]

T = TypeVar("T")
V = TypeVar("V")

UnstructureHookFactory = TypeVar(
    "UnstructureHookFactory", bound=HookFactory[UnstructureHook]
)

# The Extended factory also takes a converter.
ExtendedUnstructureHookFactory: TypeAlias = Callable[[TargetType, T], UnstructureHook]

# This typevar for the BaseConverter.
AnyUnstructureHookFactoryBase = TypeVar(
    "AnyUnstructureHookFactoryBase",
    bound="HookFactory[UnstructureHook] | ExtendedUnstructureHookFactory[BaseConverter]",
)

# This typevar for the Converter.
AnyUnstructureHookFactory = TypeVar(
    "AnyUnstructureHookFactory",
    bound="HookFactory[UnstructureHook] | ExtendedUnstructureHookFactory[Converter]",
)

StructureHookFactory = TypeVar("StructureHookFactory", bound=HookFactory[StructureHook])

# The Extended factory also takes a converter.
ExtendedStructureHookFactory: TypeAlias = Callable[[TargetType, T], StructureHook]

# This typevar for the BaseConverter.
AnyStructureHookFactoryBase = TypeVar(
    "AnyStructureHookFactoryBase",
    bound="HookFactory[StructureHook] | ExtendedStructureHookFactory[BaseConverter]",
)

# This typevar for the Converter.
AnyStructureHookFactory = TypeVar(
    "AnyStructureHookFactory",
    bound="HookFactory[StructureHook] | ExtendedStructureHookFactory[Converter]",
)

UnstructureHookT = TypeVar("UnstructureHookT", bound=UnstructureHook)
StructureHookT = TypeVar("StructureHookT", bound=StructureHook)


class UnstructureStrategy(Enum):
    """`attrs` classes unstructuring strategies."""

    AS_DICT = "asdict"
    AS_TUPLE = "astuple"


def is_literal_containing_enums(typ: type) -> bool:
    return is_literal(typ) and any(isinstance(val, Enum) for val in typ.__args__)


def _is_extended_factory(factory: Callable) -> bool:
    """Does this factory also accept a converter arg?"""
    # We use the original `inspect.signature` to not evaluate string
    # annotations.
    sig = inspect_signature(factory)
    return (
        len(sig.parameters) >= 2
        and (list(sig.parameters.values())[1]).default is Signature.empty
    )


class BaseConverter:
    """Converts between structured and unstructured data."""

    __slots__ = (
        "_unstructure_func",
        "_unstructure_attrs",
        "_structure_attrs",
        "_dict_factory",
        "_union_struct_registry",
        "_structure_func",
        "_prefer_attrib_converters",
        "detailed_validation",
        "_struct_copy_skip",
        "_unstruct_copy_skip",
    )

    def __init__(
        self,
        dict_factory: Callable[[], Any] = dict,
        unstruct_strat: UnstructureStrategy = UnstructureStrategy.AS_DICT,
        prefer_attrib_converters: bool = False,
        detailed_validation: bool = True,
        unstructure_fallback_factory: HookFactory[UnstructureHook] = lambda _: identity,
        structure_fallback_factory: HookFactory[StructureHook] = lambda _: raise_error,
    ) -> None:
        """
        :param detailed_validation: Whether to use a slightly slower mode for detailed
            validation errors.
        :param unstructure_fallback_factory: A hook factory to be called when no
            registered unstructuring hooks match.
        :param structure_fallback_factory: A hook factory to be called when no
            registered structuring hooks match.

        ..  versionadded:: 23.2.0 *unstructure_fallback_factory*
        ..  versionadded:: 23.2.0 *structure_fallback_factory*
        """
        unstruct_strat = UnstructureStrategy(unstruct_strat)
        self._prefer_attrib_converters = prefer_attrib_converters

        self.detailed_validation = detailed_validation
        self._union_struct_registry: dict[Any, Callable[[Any, type[T]], T]] = {}

        # Create a per-instance cache.
        if unstruct_strat is UnstructureStrategy.AS_DICT:
            self._unstructure_attrs = self.unstructure_attrs_asdict
            self._structure_attrs = self.structure_attrs_fromdict
        else:
            self._unstructure_attrs = self.unstructure_attrs_astuple
            self._structure_attrs = self.structure_attrs_fromtuple

        self._unstructure_func = MultiStrategyDispatch(
            unstructure_fallback_factory, self
        )
        self._unstructure_func.register_cls_list(
            [(bytes, identity), (str, identity), (Path, str)]
        )
        self._unstructure_func.register_func_list(
            [
                (
                    is_protocol,
                    lambda o: self.unstructure(o, unstructure_as=o.__class__),
                ),
                (
                    lambda t: get_final_base(t) is not None,
                    lambda t: self.get_unstructure_hook(get_final_base(t)),
                    True,
                ),
                (
                    is_type_alias,
                    lambda t: self.get_unstructure_hook(get_type_alias_base(t)),
                    True,
                ),
                (is_mapping, self._unstructure_mapping),
                (is_sequence, self._unstructure_seq),
                (is_mutable_set, self._unstructure_seq),
                (is_frozenset, self._unstructure_seq),
                (lambda t: issubclass(t, Enum), self._unstructure_enum),
                (has, self._unstructure_attrs),
                (is_union_type, self._unstructure_union),
                (lambda t: t in ANIES, self.unstructure),
            ]
        )

        # Per-instance register of to-attrs converters.
        # Singledispatch dispatches based on the first argument, so we
        # store the function and switch the arguments in self.loads.
        self._structure_func = MultiStrategyDispatch(structure_fallback_factory, self)
        self._structure_func.register_func_list(
            [
                (
                    lambda cl: cl in ANIES or cl is Optional or cl is None,
                    lambda v, _: v,
                ),
                (is_generic_attrs, self._gen_structure_generic, True),
                (lambda t: get_newtype_base(t) is not None, self._structure_newtype),
                (is_type_alias, self._find_type_alias_structure_hook, True),
                (
                    lambda t: get_final_base(t) is not None,
                    self._structure_final_factory,
                    True,
                ),
                (is_literal, self._structure_simple_literal),
                (is_literal_containing_enums, self._structure_enum_literal),
                (is_sequence, list_structure_factory, "extended"),
                (is_deque, self._structure_deque),
                (is_mutable_set, self._structure_set),
                (is_frozenset, self._structure_frozenset),
                (is_tuple, self._structure_tuple),
                (is_namedtuple, namedtuple_structure_factory, "extended"),
                (is_mapping, self._structure_dict),
                (is_supported_union, self._gen_attrs_union_structure, True),
                (is_optional, self._structure_optional),
                (
                    lambda t: is_union_type(t) and t in self._union_struct_registry,
                    self._union_struct_registry.__getitem__,
                    True,
                ),
                (has, self._structure_attrs),
            ]
        )
        # Strings are sequences.
        self._structure_func.register_cls_list(
            [
                (str, self._structure_call),
                (bytes, self._structure_call),
                (int, self._structure_call),
                (float, self._structure_call),
                (Enum, self._structure_call),
                (Path, self._structure_call),
            ]
        )

        self._dict_factory = dict_factory

        self._unstruct_copy_skip = self._unstructure_func.get_num_fns()
        self._struct_copy_skip = self._structure_func.get_num_fns()

    def unstructure(self, obj: Any, unstructure_as: Any = None) -> Any:
        return self._unstructure_func.dispatch(
            obj.__class__ if unstructure_as is None else unstructure_as
        )(obj)

    @property
    def unstruct_strat(self) -> UnstructureStrategy:
        """The default way of unstructuring ``attrs`` classes."""
        return (
            UnstructureStrategy.AS_DICT
            if self._unstructure_attrs == self.unstructure_attrs_asdict
            else UnstructureStrategy.AS_TUPLE
        )

    @overload
    def register_unstructure_hook(self, cls: UnstructureHookT) -> UnstructureHookT: ...

    @overload
    def register_unstructure_hook(self, cls: Any, func: UnstructureHook) -> None: ...

    def register_unstructure_hook(
        self, cls: Any = None, func: UnstructureHook | None = None
    ) -> Callable[[UnstructureHook]] | None:
        """Register a class-to-primitive converter function for a class.

        The converter function should take an instance of the class and return
        its Python equivalent.

        May also be used as a decorator. When used as a decorator, the first
        argument annotation from the decorated function will be used as the
        type to register the hook for.

        .. versionchanged:: 24.1.0
            This method may now be used as a decorator.
        """
        if func is None:
            # Autodetecting decorator.
            func = cls
            sig = signature(func)
            cls = next(iter(sig.parameters.values())).annotation
            self.register_unstructure_hook(cls, func)

            return func

        if attrs_has(cls):
            resolve_types(cls)
        if is_union_type(cls):
            self._unstructure_func.register_func_list([(lambda t: t == cls, func)])
        elif get_newtype_base(cls) is not None:
            # This is a newtype, so we handle it specially.
            self._unstructure_func.register_func_list([(lambda t: t is cls, func)])
        else:
            self._unstructure_func.register_cls_list([(cls, func)])
        return None

    def register_unstructure_hook_func(
        self, check_func: Predicate, func: UnstructureHook
    ) -> None:
        """Register a class-to-primitive converter function for a class, using
        a function to check if it's a match.
        """
        self._unstructure_func.register_func_list([(check_func, func)])

    @overload
    def register_unstructure_hook_factory(
        self, predicate: Predicate
    ) -> Callable[[AnyUnstructureHookFactoryBase], AnyUnstructureHookFactoryBase]: ...

    @overload
    def register_unstructure_hook_factory(
        self, predicate: Predicate, factory: UnstructureHookFactory
    ) -> UnstructureHookFactory: ...

    @overload
    def register_unstructure_hook_factory(
        self,
        predicate: Predicate,
        factory: ExtendedUnstructureHookFactory[BaseConverter],
    ) -> ExtendedUnstructureHookFactory[BaseConverter]: ...

    def register_unstructure_hook_factory(self, predicate, factory=None):
        """
        Register a hook factory for a given predicate.

        The hook factory may expose an additional required parameter. In this case,
        the current converter will be provided to the hook factory as that
        parameter.

        May also be used as a decorator.

        :param predicate: A function that, given a type, returns whether the factory
            can produce a hook for that type.
        :param factory: A callable that, given a type, produces an unstructuring
            hook for that type. This unstructuring hook will be cached.

        .. versionchanged:: 24.1.0
            This method may now be used as a decorator.
            The factory may also receive the converter as a second, required argument.
        """
        if factory is None:

            def decorator(factory):
                # Is this an extended factory (takes a converter too)?
                if _is_extended_factory(factory):
                    self._unstructure_func.register_func_list(
                        [(predicate, factory, "extended")]
                    )
                else:
                    self._unstructure_func.register_func_list(
                        [(predicate, factory, True)]
                    )

            return decorator

        self._unstructure_func.register_func_list(
            [
                (
                    predicate,
                    factory,
                    "extended" if _is_extended_factory(factory) else True,
                )
            ]
        )
        return factory

    def get_unstructure_hook(
        self, type: Any, cache_result: bool = True
    ) -> UnstructureHook:
        """Get the unstructure hook for the given type.

        This hook can be manually called, or composed with other functions
        and re-registered.

        If no hook is registered, the converter unstructure fallback factory
        will be used to produce one.

        :param cache: Whether to cache the returned hook.

        .. versionadded:: 24.1.0
        """
        return (
            self._unstructure_func.dispatch(type)
            if cache_result
            else self._unstructure_func.dispatch_without_caching(type)
        )

    @overload
    def register_structure_hook(self, cl: StructureHookT) -> StructureHookT: ...

    @overload
    def register_structure_hook(self, cl: Any, func: StructureHook) -> None: ...

    def register_structure_hook(
        self, cl: Any, func: StructureHook | None = None
    ) -> None:
        """Register a primitive-to-class converter function for a type.

        The converter function should take two arguments:
          * a Python object to be converted,
          * the type to convert to

        and return the instance of the class. The type may seem redundant, but
        is sometimes needed (for example, when dealing with generic classes).

        This method may be used as a decorator. In this case, the decorated
        hook must have a return type annotation, and this annotation will be used
        as the type for the hook.

        .. versionchanged:: 24.1.0
            This method may now be used as a decorator.
        """
        if func is None:
            # The autodetecting decorator.
            func = cl
            sig = signature(func)
            self.register_structure_hook(sig.return_annotation, func)
            return func

        if attrs_has(cl):
            resolve_types(cl)
        if is_union_type(cl):
            self._union_struct_registry[cl] = func
            self._structure_func.clear_cache()
        elif get_newtype_base(cl) is not None:
            # This is a newtype, so we handle it specially.
            self._structure_func.register_func_list([(lambda t: t is cl, func)])
        else:
            self._structure_func.register_cls_list([(cl, func)])
        return None

    def register_structure_hook_func(
        self, check_func: Predicate, func: StructureHook
    ) -> None:
        """Register a class-to-primitive converter function for a class, using
        a function to check if it's a match.
        """
        self._structure_func.register_func_list([(check_func, func)])

    @overload
    def register_structure_hook_factory(
        self, predicate: Predicate
    ) -> Callable[[AnyStructureHookFactoryBase], AnyStructureHookFactoryBase]: ...

    @overload
    def register_structure_hook_factory(
        self, predicate: Predicate, factory: StructureHookFactory
    ) -> StructureHookFactory: ...

    @overload
    def register_structure_hook_factory(
        self, predicate: Predicate, factory: ExtendedStructureHookFactory[BaseConverter]
    ) -> ExtendedStructureHookFactory[BaseConverter]: ...

    def register_structure_hook_factory(self, predicate, factory=None):
        """
        Register a hook factory for a given predicate.

        The hook factory may expose an additional required parameter. In this case,
        the current converter will be provided to the hook factory as that
        parameter.

        May also be used as a decorator.

        :param predicate: A function that, given a type, returns whether the factory
            can produce a hook for that type.
        :param factory: A callable that, given a type, produces a structuring
            hook for that type. This structuring hook will be cached.

        .. versionchanged:: 24.1.0
            This method may now be used as a decorator.
            The factory may also receive the converter as a second, required argument.
        """
        if factory is None:
            # Decorator use.
            def decorator(factory):
                # Is this an extended factory (takes a converter too)?
                if _is_extended_factory(factory):
                    self._structure_func.register_func_list(
                        [(predicate, factory, "extended")]
                    )
                else:
                    self._structure_func.register_func_list(
                        [(predicate, factory, True)]
                    )

            return decorator
        self._structure_func.register_func_list(
            [
                (
                    predicate,
                    factory,
                    "extended" if _is_extended_factory(factory) else True,
                )
            ]
        )
        return factory

    def structure(self, obj: UnstructuredValue, cl: type[T]) -> T:
        """Convert unstructured Python data structures to structured data."""
        return self._structure_func.dispatch(cl)(obj, cl)

    def get_structure_hook(self, type: Any, cache_result: bool = True) -> StructureHook:
        """Get the structure hook for the given type.

        This hook can be manually called, or composed with other functions
        and re-registered.

        If no hook is registered, the converter structure fallback factory
        will be used to produce one.

        :param cache: Whether to cache the returned hook.

        .. versionadded:: 24.1.0
        """
        return (
            self._structure_func.dispatch(type)
            if cache_result
            else self._structure_func.dispatch_without_caching(type)
        )

    # Classes to Python primitives.
    def unstructure_attrs_asdict(self, obj: Any) -> dict[str, Any]:
        """Our version of `attrs.asdict`, so we can call back to us."""
        attrs = fields(obj.__class__)
        dispatch = self._unstructure_func.dispatch
        rv = self._dict_factory()
        for a in attrs:
            name = a.name
            v = getattr(obj, name)
            rv[name] = dispatch(a.type or v.__class__)(v)
        return rv

    def unstructure_attrs_astuple(self, obj: Any) -> tuple[Any, ...]:
        """Our version of `attrs.astuple`, so we can call back to us."""
        attrs = fields(obj.__class__)
        dispatch = self._unstructure_func.dispatch
        res = []
        for a in attrs:
            name = a.name
            v = getattr(obj, name)
            res.append(dispatch(a.type or v.__class__)(v))
        return tuple(res)

    def _unstructure_enum(self, obj: Enum) -> Any:
        """Convert an enum to its value."""
        return obj.value

    def _unstructure_seq(self, seq: Sequence[T]) -> Sequence[T]:
        """Convert a sequence to primitive equivalents."""
        # We can reuse the sequence class, so tuples stay tuples.
        dispatch = self._unstructure_func.dispatch
        return seq.__class__(dispatch(e.__class__)(e) for e in seq)

    def _unstructure_mapping(self, mapping: Mapping[T, V]) -> Mapping[T, V]:
        """Convert a mapping of attr classes to primitive equivalents."""

        # We can reuse the mapping class, so dicts stay dicts and OrderedDicts
        # stay OrderedDicts.
        dispatch = self._unstructure_func.dispatch
        return mapping.__class__(
            (dispatch(k.__class__)(k), dispatch(v.__class__)(v))
            for k, v in mapping.items()
        )

    # note: Use UnionType when 3.11 is released as
    # the behaviour of @final is changed. This would
    # affect how we can support UnionType in ._compat.py
    def _unstructure_union(self, obj: Any) -> Any:
        """
        Unstructure an object as a union.

        By default, just unstructures the instance.
        """
        return self._unstructure_func.dispatch(obj.__class__)(obj)

    # Python primitives to classes.

    def _gen_structure_generic(self, cl: type[T]) -> DictStructureFn[T]:
        """Create and return a hook for structuring generics."""
        return make_dict_structure_fn(
            cl, self, _cattrs_prefer_attrib_converters=self._prefer_attrib_converters
        )

    def _gen_attrs_union_structure(
        self, cl: Any, use_literals: bool = True
    ) -> Callable[[Any, type[T]], type[T] | None]:
        """
        Generate a structuring function for a union of attrs classes (and maybe None).

        :param use_literals: Whether to consider literal fields.
        """
        dis_fn = self._get_dis_func(cl, use_literals=use_literals)
        has_none = NoneType in cl.__args__

        if has_none:

            def structure_attrs_union(obj, _) -> cl:
                if obj is None:
                    return None
                return self.structure(obj, dis_fn(obj))

        else:

            def structure_attrs_union(obj, _):
                return self.structure(obj, dis_fn(obj))

        return structure_attrs_union

    @staticmethod
    def _structure_call(obj: Any, cl: type[T]) -> Any:
        """Just call ``cl`` with the given ``obj``.

        This is just an optimization on the ``_structure_default`` case, when
        we know we can skip the ``if`` s. Use for ``str``, ``bytes``, ``enum``,
        etc.
        """
        return cl(obj)

    @staticmethod
    def _structure_simple_literal(val, type):
        if val not in type.__args__:
            raise Exception(f"{val} not in literal {type}")
        return val

    @staticmethod
    def _structure_enum_literal(val, type):
        vals = {(x.value if isinstance(x, Enum) else x): x for x in type.__args__}
        try:
            return vals[val]
        except KeyError:
            raise Exception(f"{val} not in literal {type}") from None

    def _structure_newtype(self, val: UnstructuredValue, type) -> StructuredValue:
        base = get_newtype_base(type)
        return self.get_structure_hook(base)(val, base)

    def _find_type_alias_structure_hook(self, type: Any) -> StructureHook:
        base = get_type_alias_base(type)
        res = self.get_structure_hook(base)
        if res == self._structure_call:
            # we need to replace the type arg of `structure_call`
            return lambda v, _, __base=base: __base(v)
        return lambda v, _, __base=base: res(v, __base)

    def _structure_final_factory(self, type):
        base = get_final_base(type)
        res = self.get_structure_hook(base)
        return lambda v, _, __base=base: res(v, __base)

    # Attrs classes.

    def structure_attrs_fromtuple(self, obj: tuple[Any, ...], cl: type[T]) -> T:
        """Load an attrs class from a sequence (tuple)."""
        conv_obj = []  # A list of converter parameters.
        for a, value in zip(fields(cl), obj):
            # We detect the type by the metadata.
            converted = self._structure_attribute(a, value)
            conv_obj.append(converted)

        return cl(*conv_obj)

    def _structure_attribute(self, a: Attribute | Field, value: Any) -> Any:
        """Handle an individual attrs attribute."""
        type_ = a.type
        attrib_converter = getattr(a, "converter", None)
        if self._prefer_attrib_converters and attrib_converter:
            # A attrib converter is defined on this attribute, and
            # prefer_attrib_converters is set to give these priority over registered
            # structure hooks. So, pass through the raw value, which attrs will flow
            # into the converter
            return value
        if type_ is None:
            # No type metadata.
            return value

        try:
            return self._structure_func.dispatch(type_)(value, type_)
        except StructureHandlerNotFoundError:
            if attrib_converter:
                # Return the original value and fallback to using an attrib converter.
                return value
            raise

    def structure_attrs_fromdict(self, obj: Mapping[str, Any], cl: type[T]) -> T:
        """Instantiate an attrs class from a mapping (dict)."""
        # For public use.

        conv_obj = {}  # Start with a fresh dict, to ignore extra keys.
        for a in fields(cl):
            try:
                val = obj[a.name]
            except KeyError:
                continue

            # try .alias and .name because this code also supports dataclasses!
            conv_obj[getattr(a, "alias", a.name)] = self._structure_attribute(a, val)

        return cl(**conv_obj)

    def _structure_deque(self, obj: Iterable[T], cl: Any) -> deque[T]:
        """Convert an iterable to a potentially generic deque."""
        if is_bare(cl) or cl.__args__[0] in ANIES:
            res = deque(obj)
        else:
            elem_type = cl.__args__[0]
            handler = self._structure_func.dispatch(elem_type)
            if self.detailed_validation:
                errors = []
                res = deque()
                ix = 0  # Avoid `enumerate` for performance.
                for e in obj:
                    try:
                        res.append(handler(e, elem_type))
                    except Exception as e:
                        msg = IterableValidationNote(
                            f"Structuring {cl} @ index {ix}", ix, elem_type
                        )
                        e.__notes__ = [*getattr(e, "__notes__", []), msg]
                        errors.append(e)
                    finally:
                        ix += 1
                if errors:
                    raise IterableValidationError(
                        f"While structuring {cl!r}", errors, cl
                    )
            else:
                res = deque(handler(e, elem_type) for e in obj)
        return res

    def _structure_set(
        self, obj: Iterable[T], cl: Any, structure_to: type = set
    ) -> Set[T]:
        """Convert an iterable into a potentially generic set."""
        if is_bare(cl) or cl.__args__[0] in ANIES:
            return structure_to(obj)
        elem_type = cl.__args__[0]
        handler = self._structure_func.dispatch(elem_type)
        if self.detailed_validation:
            errors = []
            res = set()
            ix = 0
            for e in obj:
                try:
                    res.add(handler(e, elem_type))
                except Exception as exc:
                    msg = IterableValidationNote(
                        f"Structuring {structure_to.__name__} @ element {e!r}",
                        ix,
                        elem_type,
                    )
                    exc.__notes__ = [*getattr(exc, "__notes__", []), msg]
                    errors.append(exc)
                finally:
                    ix += 1
            if errors:
                raise IterableValidationError(f"While structuring {cl!r}", errors, cl)
            return res if structure_to is set else structure_to(res)
        if structure_to is set:
            return {handler(e, elem_type) for e in obj}
        return structure_to([handler(e, elem_type) for e in obj])

    def _structure_frozenset(
        self, obj: Iterable[T], cl: Any
    ) -> FrozenSetSubscriptable[T]:
        """Convert an iterable into a potentially generic frozenset."""
        return self._structure_set(obj, cl, structure_to=frozenset)

    def _structure_dict(self, obj: Mapping[T, V], cl: Any) -> dict[T, V]:
        """Convert a mapping into a potentially generic dict."""
        if is_bare(cl) or cl.__args__ == (Any, Any):
            return dict(obj)
        key_type, val_type = cl.__args__

        if self.detailed_validation:
            key_handler = self._structure_func.dispatch(key_type)
            val_handler = self._structure_func.dispatch(val_type)
            errors = []
            res = {}

            for k, v in obj.items():
                try:
                    value = val_handler(v, val_type)
                except Exception as exc:
                    msg = IterableValidationNote(
                        f"Structuring mapping value @ key {k!r}", k, val_type
                    )
                    exc.__notes__ = [*getattr(exc, "__notes__", []), msg]
                    errors.append(exc)
                    continue

                try:
                    key = key_handler(k, key_type)
                    res[key] = value
                except Exception as exc:
                    msg = IterableValidationNote(
                        f"Structuring mapping key @ key {k!r}", k, key_type
                    )
                    exc.__notes__ = [*getattr(exc, "__notes__", []), msg]
                    errors.append(exc)

            if errors:
                raise IterableValidationError(f"While structuring {cl!r}", errors, cl)
            return res

        if key_type in ANIES:
            val_conv = self._structure_func.dispatch(val_type)
            return {k: val_conv(v, val_type) for k, v in obj.items()}
        if val_type in ANIES:
            key_conv = self._structure_func.dispatch(key_type)
            return {key_conv(k, key_type): v for k, v in obj.items()}
        key_conv = self._structure_func.dispatch(key_type)
        val_conv = self._structure_func.dispatch(val_type)
        return {key_conv(k, key_type): val_conv(v, val_type) for k, v in obj.items()}

    def _structure_optional(self, obj, union):
        if obj is None:
            return None
        union_params = union.__args__
        other = union_params[0] if union_params[1] is NoneType else union_params[1]
        # We can't actually have a Union of a Union, so this is safe.
        return self._structure_func.dispatch(other)(obj, other)

    def _structure_tuple(self, obj: Any, tup: type[T]) -> T:
        """Deal with structuring into a tuple."""
        tup_params = None if tup in (Tuple, tuple) else tup.__args__
        has_ellipsis = tup_params and tup_params[-1] is Ellipsis
        if tup_params is None or (has_ellipsis and tup_params[0] in ANIES):
            # Just a Tuple. (No generic information.)
            return tuple(obj)
        if has_ellipsis:
            # We're dealing with a homogenous tuple, Tuple[int, ...]
            tup_type = tup_params[0]
            conv = self._structure_func.dispatch(tup_type)
            if self.detailed_validation:
                errors = []
                res = []
                ix = 0
                for e in obj:
                    try:
                        res.append(conv(e, tup_type))
                    except Exception as exc:
                        msg = IterableValidationNote(
                            f"Structuring {tup} @ index {ix}", ix, tup_type
                        )
                        exc.__notes__ = [*getattr(exc, "__notes__", []), msg]
                        errors.append(exc)
                    finally:
                        ix += 1
                if errors:
                    raise IterableValidationError(
                        f"While structuring {tup!r}", errors, tup
                    )
                return tuple(res)
            return tuple(conv(e, tup_type) for e in obj)

        # We're dealing with a heterogenous tuple.
        exp_len = len(tup_params)
        try:
            len_obj = len(obj)
        except TypeError:
            pass  # most likely an unsized iterator, eg generator
        else:
            if len_obj > exp_len:
                exp_len = len_obj
        if self.detailed_validation:
            errors = []
            res = []
            for ix, (t, e) in enumerate(zip(tup_params, obj)):
                try:
                    conv = self._structure_func.dispatch(t)
                    res.append(conv(e, t))
                except Exception as exc:
                    msg = IterableValidationNote(
                        f"Structuring {tup} @ index {ix}", ix, t
                    )
                    exc.__notes__ = [*getattr(exc, "__notes__", []), msg]
                    errors.append(exc)
            if len(res) < exp_len:
                problem = "Not enough" if len(res) < len(tup_params) else "Too many"
                exc = ValueError(f"{problem} values in {obj!r} to structure as {tup!r}")
                msg = f"Structuring {tup}"
                exc.__notes__ = [*getattr(exc, "__notes__", []), msg]
                errors.append(exc)
            if errors:
                raise IterableValidationError(f"While structuring {tup!r}", errors, tup)
            return tuple(res)

        res = tuple(
            [self._structure_func.dispatch(t)(e, t) for t, e in zip(tup_params, obj)]
        )
        if len(res) < exp_len:
            problem = "Not enough" if len(res) < len(tup_params) else "Too many"
            raise ValueError(f"{problem} values in {obj!r} to structure as {tup!r}")
        return res

    def _get_dis_func(
        self,
        union: Any,
        use_literals: bool = True,
        overrides: dict[str, AttributeOverride] | None = None,
    ) -> Callable[[Any], type]:
        """Fetch or try creating a disambiguation function for a union."""
        union_types = union.__args__
        if NoneType in union_types:
            # We support unions of attrs classes and NoneType higher in the
            # logic.
            union_types = tuple(e for e in union_types if e is not NoneType)

        # TODO: technically both disambiguators could support TypedDicts and
        # dataclasses...
        if not all(has(get_origin(e) or e) for e in union_types):
            raise StructureHandlerNotFoundError(
                "Only unions of attrs classes supported "
                "currently. Register a structure hook manually.",
                type_=union,
            )

        return create_default_dis_func(
            self,
            *union_types,
            use_literals=use_literals,
            overrides=overrides if overrides is not None else "from_converter",
        )

    def __deepcopy__(self, _) -> BaseConverter:
        return self.copy()

    def copy(
        self,
        dict_factory: Callable[[], Any] | None = None,
        unstruct_strat: UnstructureStrategy | None = None,
        prefer_attrib_converters: bool | None = None,
        detailed_validation: bool | None = None,
    ) -> BaseConverter:
        """Create a copy of the converter, keeping all existing custom hooks.

        :param detailed_validation: Whether to use a slightly slower mode for detailed
            validation errors.
        """
        res = self.__class__(
            dict_factory if dict_factory is not None else self._dict_factory,
            (
                unstruct_strat
                if unstruct_strat is not None
                else (
                    UnstructureStrategy.AS_DICT
                    if self._unstructure_attrs == self.unstructure_attrs_asdict
                    else UnstructureStrategy.AS_TUPLE
                )
            ),
            (
                prefer_attrib_converters
                if prefer_attrib_converters is not None
                else self._prefer_attrib_converters
            ),
            (
                detailed_validation
                if detailed_validation is not None
                else self.detailed_validation
            ),
        )

        self._unstructure_func.copy_to(res._unstructure_func, self._unstruct_copy_skip)
        self._structure_func.copy_to(res._structure_func, self._struct_copy_skip)

        return res


class Converter(BaseConverter):
    """A converter which generates specialized un/structuring functions."""

    __slots__ = (
        "omit_if_default",
        "forbid_extra_keys",
        "type_overrides",
        "_unstruct_collection_overrides",
    )

    def __init__(
        self,
        dict_factory: Callable[[], Any] = dict,
        unstruct_strat: UnstructureStrategy = UnstructureStrategy.AS_DICT,
        omit_if_default: bool = False,
        forbid_extra_keys: bool = False,
        type_overrides: Mapping[type, AttributeOverride] = {},
        unstruct_collection_overrides: Mapping[type, Callable] = {},
        prefer_attrib_converters: bool = False,
        detailed_validation: bool = True,
        unstructure_fallback_factory: HookFactory[UnstructureHook] = lambda _: identity,
        structure_fallback_factory: HookFactory[StructureHook] = lambda _: raise_error,
    ):
        """
        :param detailed_validation: Whether to use a slightly slower mode for detailed
            validation errors.
        :param unstructure_fallback_factory: A hook factory to be called when no
            registered unstructuring hooks match.
        :param structure_fallback_factory: A hook factory to be called when no
            registered structuring hooks match.

        ..  versionadded:: 23.2.0 *unstructure_fallback_factory*
        ..  versionadded:: 23.2.0 *structure_fallback_factory*
        """
        super().__init__(
            dict_factory=dict_factory,
            unstruct_strat=unstruct_strat,
            prefer_attrib_converters=prefer_attrib_converters,
            detailed_validation=detailed_validation,
            unstructure_fallback_factory=unstructure_fallback_factory,
            structure_fallback_factory=structure_fallback_factory,
        )
        self.omit_if_default = omit_if_default
        self.forbid_extra_keys = forbid_extra_keys
        self.type_overrides = dict(type_overrides)

        unstruct_collection_overrides = {
            get_origin(k) or k: v for k, v in unstruct_collection_overrides.items()
        }

        self._unstruct_collection_overrides = unstruct_collection_overrides

        # Do a little post-processing magic to make things easier for users.
        co = unstruct_collection_overrides

        # abc.Set overrides, if defined, apply to abc.MutableSets and sets
        if OriginAbstractSet in co:
            if OriginMutableSet not in co:
                co[OriginMutableSet] = co[OriginAbstractSet]
                co[AbcMutableSet] = co[OriginAbstractSet]  # For 3.8 compatibility.
            if FrozenSetSubscriptable not in co:
                co[FrozenSetSubscriptable] = co[OriginAbstractSet]

        # abc.MutableSet overrrides, if defined, apply to sets
        if OriginMutableSet in co and set not in co:
            co[set] = co[OriginMutableSet]

        if FrozenSetSubscriptable in co:
            co[frozenset] = co[FrozenSetSubscriptable]  # For 3.8 compatibility.

        # abc.Sequence overrides, if defined, can apply to MutableSequences, lists and
        # tuples
        if Sequence in co:
            if MutableSequence not in co:
                co[MutableSequence] = co[Sequence]
            if tuple not in co:
                co[tuple] = co[Sequence]

        # abc.MutableSequence overrides, if defined, can apply to lists
        if MutableSequence in co:
            if list not in co:
                co[list] = co[MutableSequence]
            if deque not in co:
                co[deque] = co[MutableSequence]

        # abc.Mapping overrides, if defined, can apply to MutableMappings
        if Mapping in co and MutableMapping not in co:
            co[MutableMapping] = co[Mapping]

        # abc.MutableMapping overrides, if defined, can apply to dicts
        if MutableMapping in co and dict not in co:
            co[dict] = co[MutableMapping]

        # builtins.dict overrides, if defined, can apply to counters
        if dict in co and Counter not in co:
            co[Counter] = co[dict]

        if unstruct_strat is UnstructureStrategy.AS_DICT:
            # Override the attrs handler.
            self.register_unstructure_hook_factory(
                has_with_generic, self.gen_unstructure_attrs_fromdict
            )
            self.register_structure_hook_factory(
                has_with_generic, self.gen_structure_attrs_fromdict
            )
        self.register_unstructure_hook_factory(
            is_annotated, self.gen_unstructure_annotated
        )
        self.register_unstructure_hook_factory(
            is_hetero_tuple, self.gen_unstructure_hetero_tuple
        )
        self.register_unstructure_hook_factory(is_namedtuple)(
            namedtuple_unstructure_factory
        )
        self.register_unstructure_hook_factory(
            is_sequence, self.gen_unstructure_iterable
        )
        self.register_unstructure_hook_factory(is_mapping, self.gen_unstructure_mapping)
        self.register_unstructure_hook_factory(
            is_mutable_set,
            lambda cl: self.gen_unstructure_iterable(cl, unstructure_to=set),
        )
        self.register_unstructure_hook_factory(
            is_frozenset,
            lambda cl: self.gen_unstructure_iterable(cl, unstructure_to=frozenset),
        )
        self.register_unstructure_hook_factory(
            is_optional, self.gen_unstructure_optional
        )
        self.register_unstructure_hook_factory(
            is_typeddict, self.gen_unstructure_typeddict
        )
        self.register_unstructure_hook_factory(
            lambda t: get_newtype_base(t) is not None,
            lambda t: self.get_unstructure_hook(get_newtype_base(t)),
        )

        self.register_structure_hook_factory(is_annotated, self.gen_structure_annotated)
        self.register_structure_hook_factory(is_mapping, self.gen_structure_mapping)
        self.register_structure_hook_factory(is_counter, self.gen_structure_counter)
        self.register_structure_hook_factory(is_typeddict, self.gen_structure_typeddict)
        self.register_structure_hook_factory(
            lambda t: get_newtype_base(t) is not None, self.get_structure_newtype
        )

        # We keep these so we can more correctly copy the hooks.
        self._struct_copy_skip = self._structure_func.get_num_fns()
        self._unstruct_copy_skip = self._unstructure_func.get_num_fns()

    @overload
    def register_unstructure_hook_factory(
        self, predicate: Predicate
    ) -> Callable[[AnyUnstructureHookFactory], AnyUnstructureHookFactory]: ...

    @overload
    def register_unstructure_hook_factory(
        self, predicate: Predicate, factory: UnstructureHookFactory
    ) -> UnstructureHookFactory: ...

    @overload
    def register_unstructure_hook_factory(
        self, predicate: Predicate, factory: ExtendedUnstructureHookFactory[Converter]
    ) -> ExtendedUnstructureHookFactory[Converter]: ...

    def register_unstructure_hook_factory(self, predicate, factory=None):
        # This dummy wrapper is required due to how `@overload` works.
        return super().register_unstructure_hook_factory(predicate, factory)

    @overload
    def register_structure_hook_factory(
        self, predicate: Predicate
    ) -> Callable[[AnyStructureHookFactory], AnyStructureHookFactory]: ...

    @overload
    def register_structure_hook_factory(
        self, predicate: Predicate, factory: StructureHookFactory
    ) -> StructureHookFactory: ...

    @overload
    def register_structure_hook_factory(
        self, predicate: Predicate, factory: ExtendedStructureHookFactory[Converter]
    ) -> ExtendedStructureHookFactory[Converter]: ...

    def register_structure_hook_factory(self, predicate, factory=None):
        # This dummy wrapper is required due to how `@overload` works.
        return super().register_structure_hook_factory(predicate, factory)

    def get_structure_newtype(self, type: type[T]) -> Callable[[Any, Any], T]:
        base = get_newtype_base(type)
        handler = self.get_structure_hook(base)
        return lambda v, _: handler(v, base)

    def gen_unstructure_annotated(self, type):
        origin = type.__origin__
        return self.get_unstructure_hook(origin)

    def gen_structure_annotated(self, type) -> Callable:
        """A hook factory for annotated types."""
        origin = type.__origin__
        hook = self.get_structure_hook(origin)
        return lambda v, _: hook(v, origin)

    def gen_unstructure_typeddict(self, cl: Any) -> Callable[[dict], dict]:
        """Generate a TypedDict unstructure function.

        Also apply converter-scored modifications.
        """
        return make_typeddict_dict_unstruct_fn(cl, self)

    def gen_unstructure_attrs_fromdict(
        self, cl: type[T]
    ) -> Callable[[T], dict[str, Any]]:
        origin = get_origin(cl)
        attribs = fields(origin or cl)
        if attrs_has(cl) and any(isinstance(a.type, str) for a in attribs):
            # PEP 563 annotations - need to be resolved.
            resolve_types(cl)
        attrib_overrides = {
            a.name: self.type_overrides[a.type]
            for a in attribs
            if a.type in self.type_overrides
        }

        return make_dict_unstructure_fn(
            cl, self, _cattrs_omit_if_default=self.omit_if_default, **attrib_overrides
        )

    def gen_unstructure_optional(self, cl: type[T]) -> Callable[[T], Any]:
        """Generate an unstructuring hook for optional types."""
        union_params = cl.__args__
        other = union_params[0] if union_params[1] is NoneType else union_params[1]

        if isinstance(other, TypeVar):
            handler = self.unstructure
        else:
            handler = self.get_unstructure_hook(other)

        def unstructure_optional(val, _handler=handler):
            return None if val is None else _handler(val)

        return unstructure_optional

    def gen_structure_typeddict(self, cl: Any) -> Callable[[dict, Any], dict]:
        """Generate a TypedDict structure function.

        Also apply converter-scored modifications.
        """
        return make_typeddict_dict_struct_fn(
            cl, self, _cattrs_detailed_validation=self.detailed_validation
        )

    def gen_structure_attrs_fromdict(
        self, cl: type[T]
    ) -> Callable[[Mapping[str, Any], Any], T]:
        attribs = fields(get_origin(cl) or cl if is_generic(cl) else cl)
        if attrs_has(cl) and any(isinstance(a.type, str) for a in attribs):
            # PEP 563 annotations - need to be resolved.
            resolve_types(cl)
        attrib_overrides = {
            a.name: self.type_overrides[a.type]
            for a in attribs
            if a.type in self.type_overrides
        }
        return make_dict_structure_fn(
            cl,
            self,
            _cattrs_forbid_extra_keys=self.forbid_extra_keys,
            _cattrs_prefer_attrib_converters=self._prefer_attrib_converters,
            _cattrs_detailed_validation=self.detailed_validation,
            **attrib_overrides,
        )

    def gen_unstructure_iterable(
        self, cl: Any, unstructure_to: Any = None
    ) -> IterableUnstructureFn:
        unstructure_to = self._unstruct_collection_overrides.get(
            get_origin(cl) or cl, unstructure_to or list
        )
        h = iterable_unstructure_factory(cl, self, unstructure_to=unstructure_to)
        self._unstructure_func.register_cls_list([(cl, h)], direct=True)
        return h

    def gen_unstructure_hetero_tuple(
        self, cl: Any, unstructure_to: Any = None
    ) -> HeteroTupleUnstructureFn:
        unstructure_to = self._unstruct_collection_overrides.get(
            get_origin(cl) or cl, unstructure_to or tuple
        )
        h = make_hetero_tuple_unstructure_fn(cl, self, unstructure_to=unstructure_to)
        self._unstructure_func.register_cls_list([(cl, h)], direct=True)
        return h

    def gen_unstructure_mapping(
        self,
        cl: Any,
        unstructure_to: Any = None,
        key_handler: Callable[[Any, Any | None], Any] | None = None,
    ) -> MappingUnstructureFn:
        unstructure_to = self._unstruct_collection_overrides.get(
            get_origin(cl) or cl, unstructure_to or dict
        )
        h = make_mapping_unstructure_fn(
            cl, self, unstructure_to=unstructure_to, key_handler=key_handler
        )
        self._unstructure_func.register_cls_list([(cl, h)], direct=True)
        return h

    def gen_structure_counter(self, cl: Any) -> MappingStructureFn[T]:
        h = make_mapping_structure_fn(
            cl,
            self,
            structure_to=Counter,
            val_type=int,
            detailed_validation=self.detailed_validation,
        )
        self._structure_func.register_cls_list([(cl, h)], direct=True)
        return h

    def gen_structure_mapping(self, cl: Any) -> MappingStructureFn[T]:
        structure_to = get_origin(cl) or cl
        if structure_to in (
            MutableMapping,
            AbcMutableMapping,
            Mapping,
            AbcMapping,
        ):  # These default to dicts
            structure_to = dict
        h = make_mapping_structure_fn(
            cl, self, structure_to, detailed_validation=self.detailed_validation
        )
        self._structure_func.register_cls_list([(cl, h)], direct=True)
        return h

    def copy(
        self,
        dict_factory: Callable[[], Any] | None = None,
        unstruct_strat: UnstructureStrategy | None = None,
        omit_if_default: bool | None = None,
        forbid_extra_keys: bool | None = None,
        type_overrides: Mapping[type, AttributeOverride] | None = None,
        unstruct_collection_overrides: Mapping[type, Callable] | None = None,
        prefer_attrib_converters: bool | None = None,
        detailed_validation: bool | None = None,
    ) -> Converter:
        """Create a copy of the converter, keeping all existing custom hooks.

        :param detailed_validation: Whether to use a slightly slower mode for detailed
            validation errors.
        """
        res = self.__class__(
            dict_factory if dict_factory is not None else self._dict_factory,
            (
                unstruct_strat
                if unstruct_strat is not None
                else (
                    UnstructureStrategy.AS_DICT
                    if self._unstructure_attrs == self.unstructure_attrs_asdict
                    else UnstructureStrategy.AS_TUPLE
                )
            ),
            omit_if_default if omit_if_default is not None else self.omit_if_default,
            (
                forbid_extra_keys
                if forbid_extra_keys is not None
                else self.forbid_extra_keys
            ),
            type_overrides if type_overrides is not None else self.type_overrides,
            (
                unstruct_collection_overrides
                if unstruct_collection_overrides is not None
                else self._unstruct_collection_overrides
            ),
            (
                prefer_attrib_converters
                if prefer_attrib_converters is not None
                else self._prefer_attrib_converters
            ),
            (
                detailed_validation
                if detailed_validation is not None
                else self.detailed_validation
            ),
        )

        self._unstructure_func.copy_to(
            res._unstructure_func, skip=self._unstruct_copy_skip
        )
        self._structure_func.copy_to(res._structure_func, skip=self._struct_copy_skip)

        return res


GenConverter: TypeAlias = Converter



import datetime
from _operator import _SupportsComparison
from _typeshed import Incomplete, SupportsKeysAndGetItem
from collections.abc import Callable, Iterable, Mapping
from logging import Logger
from typing import ClassVar, TypeVar, overload, type_check_only
from typing_extensions import Self

_RangeMapKT = TypeVar("_RangeMapKT", bound=_SupportsComparison)

_T = TypeVar("_T")
_VT = TypeVar("_VT")

log: Logger

class _SimpleStruct:
    def __init__(self, *args, **kw) -> None: ...
    def field_names(self) -> list[str]: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...

class SYSTEMTIME(_SimpleStruct): ...
class TIME_ZONE_INFORMATION(_SimpleStruct): ...
class DYNAMIC_TIME_ZONE_INFORMATION(_SimpleStruct): ...

class TimeZoneDefinition(DYNAMIC_TIME_ZONE_INFORMATION):
    def __init__(self, *args, **kwargs) -> None: ...
    # TIME_ZONE_INFORMATION fields as obtained by __getattribute__
    bias: datetime.timedelta
    standard_name: str
    standard_start: SYSTEMTIME
    standard_bias: datetime.timedelta
    daylight_name: str
    daylight_start: SYSTEMTIME
    daylight_bias: datetime.timedelta
    def __getattribute__(self, attr: str): ...
    @classmethod
    def current(cls) -> tuple[int, Self]: ...
    def set(self) -> None: ...
    def copy(self) -> Self: ...
    def locate_daylight_start(self, year) -> datetime.datetime: ...
    def locate_standard_start(self, year) -> datetime.datetime: ...

class TimeZoneInfo(datetime.tzinfo):
    tzRegKey: ClassVar[str]
    timeZoneName: str
    fixedStandardTime: bool
    def __init__(self, param: str | TimeZoneDefinition, fix_standard_time: bool = False) -> None: ...
    @overload  # type: ignore[override] # Split definition into overrides
    def tzname(self, dt: datetime.datetime) -> str: ...
    @overload
    def tzname(self, dt: None) -> None: ...
    def getWinInfo(self, targetYear: int) -> TimeZoneDefinition: ...
    @overload  # type: ignore[override] # False-positive, our overload covers all base types
    def utcoffset(self, dt: None) -> None: ...
    @overload
    def utcoffset(self, dt: datetime.datetime) -> datetime.timedelta: ...
    @overload  # type: ignore[override] # False-positive, our overload covers all base types
    def dst(self, dt: None) -> None: ...
    @overload
    def dst(self, dt: datetime.datetime) -> datetime.timedelta: ...
    def GetDSTStartTime(self, year: int) -> datetime.datetime: ...
    def GetDSTEndTime(self, year: int) -> datetime.datetime: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    @classmethod
    def local(cls) -> Self: ...
    @classmethod
    def utc(cls) -> Self: ...
    @staticmethod
    def get_sorted_time_zone_names() -> list[str]: ...
    @staticmethod
    def get_all_time_zones() -> list[TimeZoneInfo]: ...
    @staticmethod
    def get_sorted_time_zones(key: Incomplete | None = ...): ...

def utcnow() -> datetime.datetime: ...
def now() -> datetime.datetime: ...
def GetTZCapabilities() -> dict[str, bool]: ...

class DLLHandleCache:
    def __getitem__(self, filename: str) -> int: ...

DLLCache: DLLHandleCache

def resolveMUITimeZone(spec: str) -> str | None: ...

class RangeMap(dict[_RangeMapKT, _VT]):
    sort_params: Mapping[str, Incomplete]
    match: Callable[[_RangeMapKT, _RangeMapKT], bool]
    def __init__(
        self,
        source: SupportsKeysAndGetItem[_RangeMapKT, _VT] | Iterable[tuple[_RangeMapKT, _VT]],
        sort_params: Mapping[str, Incomplete] = {},
        key_match_comparator: Callable[[_RangeMapKT, _RangeMapKT], bool] = ...,
    ) -> None: ...
    @classmethod
    def left(cls, source: SupportsKeysAndGetItem[_RangeMapKT, _VT] | Iterable[tuple[_RangeMapKT, _VT]]) -> Self: ...
    def __getitem__(self, item: _RangeMapKT) -> _VT: ...
    @overload  # type: ignore[override] # Signature simplified over dict and Mapping
    def get(self, key: _RangeMapKT, default: _T) -> _VT | _T: ...
    @overload
    def get(self, key: _RangeMapKT, default: None = None) -> _VT | None: ...
    def bounds(self) -> tuple[_RangeMapKT, _RangeMapKT]: ...
    @type_check_only
    class RangeValueUndefined: ...

    undefined_value: RangeValueUndefined

    class Item(int): ...
    first_item: Item
    last_item: Item



from _typeshed import Incomplete

from authlib.common.errors import AuthlibHTTPError

def invalid_error_characters(text: str) -> list[str]: ...

class OAuth2Error(AuthlibHTTPError):
    state: Incomplete
    redirect_uri: Incomplete
    redirect_fragment: Incomplete
    def __init__(
        self,
        description=None,
        uri=None,
        status_code=None,
        state=None,
        redirect_uri=None,
        redirect_fragment: bool = False,
        error=None,
    ) -> None: ...
    def get_body(self): ...
    def __call__(self, uri=None): ...



from _typeshed import Incomplete, Unused
from typing import ClassVar, Literal
from typing_extensions import TypeAlias

from openpyxl.chart.axis import NumericAxis, TextAxis
from openpyxl.chart.label import DataLabelList
from openpyxl.descriptors.base import Alias, Typed, _ConvertibleToBool
from openpyxl.descriptors.excel import ExtensionList
from openpyxl.descriptors.nested import NestedBool, NestedSet

from ..xml._functions_overloads import _HasTagAndGet
from ._chart import ChartBase

_RadarChartRadarStyle: TypeAlias = Literal["standard", "marker", "filled"]

class RadarChart(ChartBase):
    tagname: ClassVar[str]
    radarStyle: NestedSet[_RadarChartRadarStyle]
    type: Alias
    varyColors: NestedBool[Literal[True]]
    ser: Incomplete
    dLbls: Typed[DataLabelList, Literal[True]]
    dataLabels: Alias
    extLst: Typed[ExtensionList, Literal[True]]
    x_axis: Typed[TextAxis, Literal[False]]
    y_axis: Typed[NumericAxis, Literal[False]]
    __elements__: ClassVar[tuple[str, ...]]
    def __init__(
        self,
        radarStyle: _HasTagAndGet[_RadarChartRadarStyle] | _RadarChartRadarStyle = "standard",
        varyColors: _HasTagAndGet[_ConvertibleToBool | None] | _ConvertibleToBool | None = None,
        ser=(),
        dLbls: DataLabelList | None = None,
        extLst: Unused = None,
        **kw,
    ) -> None: ...



from _typeshed import FileDescriptorOrPath, StrOrBytesPath
from contextlib import AbstractContextManager

def temporary_chdir(new_dir: FileDescriptorOrPath) -> AbstractContextManager[None]: ...
def get_file_directory() -> str: ...
def get_temp_directory() -> str: ...
def get_themes_directory(theme_name: str | None = None, png: bool = False) -> str: ...
def create_directory(directory: StrOrBytesPath) -> StrOrBytesPath: ...



from _typeshed import Incomplete, Unused
from collections.abc import Sequence
from contextlib import AbstractContextManager
from typing import TypedDict, type_check_only
from typing_extensions import NotRequired

import gdb

from .sources import Source

@type_check_only
class _SourceBreakpoint(TypedDict):
    source: str
    line: int
    condition: NotRequired[str | None]
    hitCondition: NotRequired[str | None]
    logMessage: NotRequired[str | None]

@type_check_only
class _ExceptionFilterOptions(TypedDict):
    filderId: str
    condition: NotRequired[str | None]

@type_check_only
class _BreakpointDescriptor(TypedDict):
    id: int
    verified: bool
    reason: NotRequired[str]  # only present when verified is False. Possibly only literal "pending" or "failed"
    message: NotRequired[str]  # only present when reason == "failed"
    source: NotRequired[Source]
    line: NotRequired[int]
    instructionReference: NotRequired[str]

@type_check_only
class _SetBreakpointResult(TypedDict):
    breakpoints: list[_BreakpointDescriptor]

# frozenset entries are tuples from _SourceBreakpoint.items() or _ExceptionFilterOptions.items()
breakpoint_map: dict[str, dict[frozenset[Incomplete], gdb.Breakpoint]]

def suppress_new_breakpoint_event() -> AbstractContextManager[None]: ...
def set_breakpoint(*, source: Source, breakpoints: Sequence[_SourceBreakpoint] = (), **args: Unused) -> _SetBreakpointResult: ...
def set_fn_breakpoint(*, breakpoints: Sequence[_SourceBreakpoint], **args: Unused) -> _SetBreakpointResult: ...
def set_insn_breakpoints(
    *, breakpoints: Sequence[_SourceBreakpoint], offset: int | None = None, **args: Unused
) -> _SetBreakpointResult: ...
def set_exception_breakpoints(
    *, filters: Sequence[str], filterOptions: Sequence[_ExceptionFilterOptions] = (), **args: Unused
) -> _SetBreakpointResult: ...



from collections.abc import Iterable

regex: str

def expand_languages(languages: Iterable[str] | None = None) -> list[str]: ...
def update(language: str | None = None) -> None: ...

langs: list[str]



from collections.abc import Iterable, Iterator
from typing import NoReturn
from typing_extensions import Self, TypeAlias

from netaddr.ip import IPAddress, IPNetwork, IPRange, _IPNetworkAddr

_IPIterable: TypeAlias = IPNetwork | IPRange | IPSet | Iterable[_IPNetworkAddr | IPRange | int]

class IPSet:
    __slots__ = ("_cidrs", "__weakref__")
    def __init__(self, iterable: _IPIterable | None = None, flags: int = 0) -> None: ...
    def compact(self) -> None: ...
    def __hash__(self) -> NoReturn: ...
    def __contains__(self, ip: _IPNetworkAddr) -> bool: ...
    def __bool__(self) -> bool: ...
    def __iter__(self) -> Iterator[IPAddress]: ...
    def iter_cidrs(self) -> list[IPNetwork]: ...
    def add(self, addr: IPRange | _IPNetworkAddr | int, flags: int = 0) -> None: ...
    def remove(self, addr: IPRange | _IPNetworkAddr | int, flags: int = 0) -> None: ...
    def pop(self) -> IPNetwork: ...
    def isdisjoint(self, other: IPSet) -> bool: ...
    def copy(self) -> Self: ...
    def update(self, iterable: _IPIterable, flags: int = 0) -> None: ...
    def clear(self) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __ne__(self, other: object) -> bool: ...
    def __lt__(self, other: IPSet) -> bool: ...
    def issubset(self, other: IPSet) -> bool: ...
    __le__ = issubset
    def __gt__(self, other: IPSet) -> bool: ...
    def issuperset(self, other: IPSet) -> bool: ...
    __ge__ = issuperset
    def union(self, other: IPSet) -> Self: ...
    __or__ = union
    def intersection(self, other: IPSet) -> IPSet: ...
    __and__ = intersection
    def symmetric_difference(self, other: IPSet) -> IPSet: ...
    __xor__ = symmetric_difference
    def difference(self, other: IPSet) -> IPSet: ...
    __sub__ = difference
    def __len__(self) -> int: ...
    @property
    def size(self) -> int: ...
    def iscontiguous(self) -> bool: ...
    def iprange(self) -> IPRange | None: ...
    def iter_ipranges(self) -> Iterator[IPRange]: ...



from logging import Logger
from typing import Final

ASSERTION_TYPE: Final[str]
log: Logger

class JWTBearerClientAssertion:
    CLIENT_ASSERTION_TYPE: Final[str]
    CLIENT_AUTH_METHOD: Final[str]
    token_url: str
    leeway: int
    def __init__(self, token_url: str, validate_jti: bool = True, leeway: int = 60) -> None: ...
    def __call__(self, query_client, request): ...
    def create_claims_options(self): ...
    def process_assertion_claims(self, assertion, resolve_key): ...
    def authenticate_client(self, client): ...
    def create_resolve_key_func(self, query_client, request): ...
    def validate_jti(self, claims, jti) -> None: ...
    def resolve_client_public_key(self, client, headers) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2017 The TensorFlow Authors. All Rights Reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
==============================================================================
"""

import builtins
import collections.abc
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class DeviceProperties(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class EnvironmentEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        value: builtins.str
        def __init__(self, *, key: builtins.str | None = ..., value: builtins.str | None = ...) -> None: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    TYPE_FIELD_NUMBER: builtins.int
    VENDOR_FIELD_NUMBER: builtins.int
    MODEL_FIELD_NUMBER: builtins.int
    FREQUENCY_FIELD_NUMBER: builtins.int
    NUM_CORES_FIELD_NUMBER: builtins.int
    ENVIRONMENT_FIELD_NUMBER: builtins.int
    NUM_REGISTERS_FIELD_NUMBER: builtins.int
    L1_CACHE_SIZE_FIELD_NUMBER: builtins.int
    L2_CACHE_SIZE_FIELD_NUMBER: builtins.int
    L3_CACHE_SIZE_FIELD_NUMBER: builtins.int
    SHARED_MEMORY_SIZE_PER_MULTIPROCESSOR_FIELD_NUMBER: builtins.int
    MEMORY_SIZE_FIELD_NUMBER: builtins.int
    BANDWIDTH_FIELD_NUMBER: builtins.int
    type: builtins.str
    """Device type (CPU, GPU, ...)"""
    vendor: builtins.str
    """Vendor (Intel, nvidia, ...)"""
    model: builtins.str
    """Model (Haswell, K40, ...)"""
    frequency: builtins.int
    """Core Frequency in Mhz"""
    num_cores: builtins.int
    """Number of cores"""
    num_registers: builtins.int
    """Number of registers per core."""
    l1_cache_size: builtins.int
    """L1 cache size in bytes"""
    l2_cache_size: builtins.int
    """L2 cache size in bytes"""
    l3_cache_size: builtins.int
    """L3 cache size in bytes"""
    shared_memory_size_per_multiprocessor: builtins.int
    """Shared memory size per multiprocessor in bytes. This field is
    applicable to GPUs only.
    """
    memory_size: builtins.int
    """Memory size in bytes"""
    bandwidth: builtins.int
    """Memory bandwidth in KB/s"""
    @property
    def environment(self) -> google.protobuf.internal.containers.ScalarMap[builtins.str, builtins.str]:
        """Version of the tools and libraries used with this device (e.g. gcc 4.9,
        cudnn 5.1)
        """

    def __init__(
        self,
        *,
        type: builtins.str | None = ...,
        vendor: builtins.str | None = ...,
        model: builtins.str | None = ...,
        frequency: builtins.int | None = ...,
        num_cores: builtins.int | None = ...,
        environment: collections.abc.Mapping[builtins.str, builtins.str] | None = ...,
        num_registers: builtins.int | None = ...,
        l1_cache_size: builtins.int | None = ...,
        l2_cache_size: builtins.int | None = ...,
        l3_cache_size: builtins.int | None = ...,
        shared_memory_size_per_multiprocessor: builtins.int | None = ...,
        memory_size: builtins.int | None = ...,
        bandwidth: builtins.int | None = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "bandwidth",
            b"bandwidth",
            "environment",
            b"environment",
            "frequency",
            b"frequency",
            "l1_cache_size",
            b"l1_cache_size",
            "l2_cache_size",
            b"l2_cache_size",
            "l3_cache_size",
            b"l3_cache_size",
            "memory_size",
            b"memory_size",
            "model",
            b"model",
            "num_cores",
            b"num_cores",
            "num_registers",
            b"num_registers",
            "shared_memory_size_per_multiprocessor",
            b"shared_memory_size_per_multiprocessor",
            "type",
            b"type",
            "vendor",
            b"vendor",
        ],
    ) -> None: ...

global___DeviceProperties = DeviceProperties

@typing.final
class NamedDevice(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    NAME_FIELD_NUMBER: builtins.int
    PROPERTIES_FIELD_NUMBER: builtins.int
    name: builtins.str
    @property
    def properties(self) -> global___DeviceProperties: ...
    def __init__(self, *, name: builtins.str | None = ..., properties: global___DeviceProperties | None = ...) -> None: ...
    def HasField(self, field_name: typing.Literal["properties", b"properties"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["name", b"name", "properties", b"properties"]) -> None: ...

global___NamedDevice = NamedDevice



from _typeshed import Incomplete

whitelist: Incomplete

def inject_header(wrapped, instance, args, kwargs): ...
def aws_meta_processor(wrapped, instance, args, kwargs, return_value, exception, subsegment, stack) -> None: ...



from __future__ import annotations

import importlib.util
import pathlib
import sys
import zipfile
from collections.abc import Sequence
from importlib.machinery import ModuleSpec
from types import ModuleType
from typing_extensions import Self

if sys.version_info >= (3, 11):
    from importlib.resources.abc import Traversable
else:
    from importlib.abc import Traversable


# Assert that some Path classes are Traversable.
def traverse(t: Traversable) -> None:
    pass


traverse(pathlib.Path())
traverse(zipfile.Path(""))


class MetaFinder:
    @classmethod
    def find_spec(cls, fullname: str, path: Sequence[str] | None, target: ModuleType | None = None) -> ModuleSpec | None:
        return None  # simplified mock for demonstration purposes only


class PathFinder:
    @classmethod
    def path_hook(cls, path_entry: str) -> type[Self]:
        return cls  # simplified mock for demonstration purposes only

    @classmethod
    def find_spec(cls, fullname: str, target: ModuleType | None = None) -> ModuleSpec | None:
        return None  # simplified mock for demonstration purposes only


class Loader:
    @classmethod
    def load_module(cls, fullname: str) -> ModuleType:
        return ModuleType(fullname)


sys.meta_path.append(MetaFinder)
sys.path_hooks.append(PathFinder.path_hook)
importlib.util.spec_from_loader("xxxx42xxxx", Loader)



import contextlib
import enum
import sys
from collections.abc import Callable, Generator, Iterable, Iterator
from typing import Any, ClassVar, cast
from typing_extensions import Self

from pynput._util import AbstractListener

class KeyCode:
    _PLATFORM_EXTENSIONS: ClassVar[Iterable[str]]  # undocumented
    vk: int | None
    char: str | None
    is_dead: bool | None
    combining: str | None
    def __init__(self, vk: str | None = None, char: str | None = None, is_dead: bool = False, **kwargs: str) -> None: ...
    def __eq__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def join(self, key: Self) -> Self: ...
    @classmethod
    def from_vk(cls, vk: int, **kwargs: Any) -> Self: ...
    @classmethod
    def from_char(cls, char: str, **kwargs: Any) -> Self: ...
    @classmethod
    def from_dead(cls, char: str, **kwargs: Any) -> Self: ...

class Key(enum.Enum):
    alt = cast(KeyCode, ...)
    alt_l = cast(KeyCode, ...)
    alt_r = cast(KeyCode, ...)
    alt_gr = cast(KeyCode, ...)
    backspace = cast(KeyCode, ...)
    caps_lock = cast(KeyCode, ...)
    cmd = cast(KeyCode, ...)
    cmd_l = cast(KeyCode, ...)
    cmd_r = cast(KeyCode, ...)
    ctrl = cast(KeyCode, ...)
    ctrl_l = cast(KeyCode, ...)
    ctrl_r = cast(KeyCode, ...)
    delete = cast(KeyCode, ...)
    down = cast(KeyCode, ...)
    end = cast(KeyCode, ...)
    enter = cast(KeyCode, ...)
    esc = cast(KeyCode, ...)
    f1 = cast(KeyCode, ...)
    f2 = cast(KeyCode, ...)
    f3 = cast(KeyCode, ...)
    f4 = cast(KeyCode, ...)
    f5 = cast(KeyCode, ...)
    f6 = cast(KeyCode, ...)
    f7 = cast(KeyCode, ...)
    f8 = cast(KeyCode, ...)
    f9 = cast(KeyCode, ...)
    f10 = cast(KeyCode, ...)
    f11 = cast(KeyCode, ...)
    f12 = cast(KeyCode, ...)
    f13 = cast(KeyCode, ...)
    f14 = cast(KeyCode, ...)
    f15 = cast(KeyCode, ...)
    f16 = cast(KeyCode, ...)
    f17 = cast(KeyCode, ...)
    f18 = cast(KeyCode, ...)
    f19 = cast(KeyCode, ...)
    f20 = cast(KeyCode, ...)
    if sys.platform == "win32":
        f21 = cast(KeyCode, ...)
        f22 = cast(KeyCode, ...)
        f23 = cast(KeyCode, ...)
        f24 = cast(KeyCode, ...)
    home = cast(KeyCode, ...)
    left = cast(KeyCode, ...)
    page_down = cast(KeyCode, ...)
    page_up = cast(KeyCode, ...)
    right = cast(KeyCode, ...)
    shift = cast(KeyCode, ...)
    shift_l = cast(KeyCode, ...)
    shift_r = cast(KeyCode, ...)
    space = cast(KeyCode, ...)
    tab = cast(KeyCode, ...)
    up = cast(KeyCode, ...)
    media_play_pause = cast(KeyCode, ...)
    media_stop = cast(KeyCode, ...)
    media_volume_mute = cast(KeyCode, ...)
    media_volume_down = cast(KeyCode, ...)
    media_volume_up = cast(KeyCode, ...)
    media_previous = cast(KeyCode, ...)
    media_next = cast(KeyCode, ...)
    if sys.platform == "darwin":
        media_eject = cast(KeyCode, ...)
    insert = cast(KeyCode, ...)
    menu = cast(KeyCode, ...)
    num_lock = cast(KeyCode, ...)
    pause = cast(KeyCode, ...)
    print_screen = cast(KeyCode, ...)
    scroll_lock = cast(KeyCode, ...)

class Controller:
    _KeyCode: ClassVar[type[KeyCode]]  # undocumented
    _Key: ClassVar[type[Key]]  # undocumented

    if sys.platform == "linux":
        CTRL_MASK: ClassVar[int]
        SHIFT_MASK: ClassVar[int]

    class InvalidKeyException(Exception): ...
    class InvalidCharacterException(Exception): ...

    def __init__(self) -> None: ...
    def press(self, key: str | Key | KeyCode) -> None: ...
    def release(self, key: str | Key | KeyCode) -> None: ...
    def tap(self, key: str | Key | KeyCode) -> None: ...
    def touch(self, key: str | Key | KeyCode, is_press: bool) -> None: ...
    @contextlib.contextmanager
    def pressed(self, *args: str | Key | KeyCode) -> Generator[None]: ...
    def type(self, string: str) -> None: ...
    @property
    def modifiers(self) -> contextlib.AbstractContextManager[Iterator[set[Key]]]: ...
    @property
    def alt_pressed(self) -> bool: ...
    @property
    def alt_gr_pressed(self) -> bool: ...
    @property
    def ctrl_pressed(self) -> bool: ...
    @property
    def shift_pressed(self) -> bool: ...

class Listener(AbstractListener):
    def __init__(
        self,
        on_press: Callable[[Key | KeyCode | None], None] | None = None,
        on_release: Callable[[Key | KeyCode | None], None] | None = None,
        suppress: bool = False,
        **kwargs: Any,
    ) -> None: ...
    def canonical(self, key: Key | KeyCode) -> Key | KeyCode: ...



from email._policybase import Policy
from email.mime.nonmultipart import MIMENonMultipart

__all__ = ["MIMEText"]

class MIMEText(MIMENonMultipart):
    def __init__(
        self, _text: str, _subtype: str = "plain", _charset: str | None = None, *, policy: Policy | None = None
    ) -> None: ...



from __future__ import annotations

from typing import Any, cast
from typing_extensions import assert_type

import grpc


class Request:
    pass


class Response:
    pass


def unary_unary_call(rq: Request, ctx: grpc.ServicerContext) -> Response:
    assert_type(rq, Request)
    return Response()


class ServiceHandler(grpc.ServiceRpcHandler):
    def service_name(self) -> str:
        return "hello"

    def service(self, handler_call_details: grpc.HandlerCallDetails) -> grpc.RpcMethodHandler[Any, Any] | None:
        rpc = grpc.RpcMethodHandler[Request, Response]()
        rpc.unary_unary = unary_unary_call
        return rpc


h = ServiceHandler()
ctx = cast(grpc.ServicerContext, None)
svc = h.service(grpc.HandlerCallDetails())
if svc is not None and svc.unary_unary is not None:
    svc.unary_unary(Request(), ctx)



import ast

from flake8_simplify.utils import UnaryOp

def get_sim201(node: UnaryOp) -> list[tuple[int, int, str]]: ...
def get_sim202(node: UnaryOp) -> list[tuple[int, int, str]]: ...
def get_sim203(node: UnaryOp) -> list[tuple[int, int, str]]: ...
def get_sim208(node: ast.UnaryOp) -> list[tuple[int, int, str]]: ...



from .auth import ClientAuth as ClientAuth, TokenAuth as TokenAuth
from .base import OAuth2Error as OAuth2Error
from .client import OAuth2Client as OAuth2Client
from .rfc6749 import (
    AuthorizationServer as AuthorizationServer,
    ClientAuthentication as ClientAuthentication,
    JsonRequest as JsonRequest,
    OAuth2Request as OAuth2Request,
    ResourceProtector as ResourceProtector,
)

__all__ = [
    "OAuth2Error",
    "ClientAuth",
    "TokenAuth",
    "OAuth2Client",
    "OAuth2Request",
    "JsonRequest",
    "AuthorizationServer",
    "ClientAuthentication",
    "ResourceProtector",
]



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import sys
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import tensorflow.core.framework.graph_debug_info_pb2
import tensorflow.core.framework.tensor_pb2

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class _TensorDebugMode:
    ValueType = typing.NewType("ValueType", builtins.int)
    V: typing_extensions.TypeAlias = ValueType

class _TensorDebugModeEnumTypeWrapper(
    google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_TensorDebugMode.ValueType], builtins.type
):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
    UNSPECIFIED: _TensorDebugMode.ValueType  # 0
    NO_TENSOR: _TensorDebugMode.ValueType  # 1
    """Only records what tensors are computed, eagerly or in graphs.
    No information regarding the value of the tensor is available.
    """
    CURT_HEALTH: _TensorDebugMode.ValueType  # 2
    """A minimalist health summary for float-type tensors.
    Contains information only about the presence/absence of pathological
    values including Infinity and NaN.
    Applicable only to float dtypes.
    """
    CONCISE_HEALTH: _TensorDebugMode.ValueType  # 3
    """A concise health summary for float-type tensors.
    Contains more information that CURT_HEALTH.
    Infinity and NaN are treated differently.
    Applicable only to float and integer dtypes.
    """
    FULL_HEALTH: _TensorDebugMode.ValueType  # 4
    """A detailed health summary.
    Contains further detailed information than `CONCISE_HEALTH`.
    Information about device, dtype and shape are included.
    Counts for various types of values (Infinity, NaN, negative, zero,
    positive) are included.
    Applicable to float, integer and boolean dtypes.
    """
    SHAPE: _TensorDebugMode.ValueType  # 5
    """Provides full runtime shape information, up to a maximum rank, beyond
    which the dimension sizes are truncated.
    """
    FULL_NUMERICS: _TensorDebugMode.ValueType  # 6
    """Full numeric summary.
    Including device, dtype, shape, counts of various types of values
    (Infinity, NaN, negative, zero, positive), and summary statistics
    (minimum, maximum, mean and variance).
    Applicable to float, integer and boolean dtypes.
    """
    FULL_TENSOR: _TensorDebugMode.ValueType  # 7
    """Full tensor value."""
    REDUCE_INF_NAN_THREE_SLOTS: _TensorDebugMode.ValueType  # 8
    """Reduce the elements of a tensor to a rank-1 tensor of shape [3], in which
    - the 1st element is -inf if any element of the tensor is -inf,
      or zero otherwise.
    - the 2nd element is +inf if any element of the tensor is +inf,
      or zero otherwise.
    - the 3rd element is nan if any element of the tensor is nan, or zero
      otherwise.
    """

class TensorDebugMode(_TensorDebugMode, metaclass=_TensorDebugModeEnumTypeWrapper):
    """Available modes for extracting debugging information from a Tensor.
    TODO(cais): Document the detailed column names and semantics in a separate
    markdown file once the implementation settles.
    """

UNSPECIFIED: TensorDebugMode.ValueType  # 0
NO_TENSOR: TensorDebugMode.ValueType  # 1
"""Only records what tensors are computed, eagerly or in graphs.
No information regarding the value of the tensor is available.
"""
CURT_HEALTH: TensorDebugMode.ValueType  # 2
"""A minimalist health summary for float-type tensors.
Contains information only about the presence/absence of pathological
values including Infinity and NaN.
Applicable only to float dtypes.
"""
CONCISE_HEALTH: TensorDebugMode.ValueType  # 3
"""A concise health summary for float-type tensors.
Contains more information that CURT_HEALTH.
Infinity and NaN are treated differently.
Applicable only to float and integer dtypes.
"""
FULL_HEALTH: TensorDebugMode.ValueType  # 4
"""A detailed health summary.
Contains further detailed information than `CONCISE_HEALTH`.
Information about device, dtype and shape are included.
Counts for various types of values (Infinity, NaN, negative, zero,
positive) are included.
Applicable to float, integer and boolean dtypes.
"""
SHAPE: TensorDebugMode.ValueType  # 5
"""Provides full runtime shape information, up to a maximum rank, beyond
which the dimension sizes are truncated.
"""
FULL_NUMERICS: TensorDebugMode.ValueType  # 6
"""Full numeric summary.
Including device, dtype, shape, counts of various types of values
(Infinity, NaN, negative, zero, positive), and summary statistics
(minimum, maximum, mean and variance).
Applicable to float, integer and boolean dtypes.
"""
FULL_TENSOR: TensorDebugMode.ValueType  # 7
"""Full tensor value."""
REDUCE_INF_NAN_THREE_SLOTS: TensorDebugMode.ValueType  # 8
"""Reduce the elements of a tensor to a rank-1 tensor of shape [3], in which
- the 1st element is -inf if any element of the tensor is -inf,
  or zero otherwise.
- the 2nd element is +inf if any element of the tensor is +inf,
  or zero otherwise.
- the 3rd element is nan if any element of the tensor is nan, or zero
  otherwise.
"""
global___TensorDebugMode = TensorDebugMode

@typing.final
class DebugEvent(google.protobuf.message.Message):
    """An Event related to the debugging of a TensorFlow program."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    WALL_TIME_FIELD_NUMBER: builtins.int
    STEP_FIELD_NUMBER: builtins.int
    DEBUG_METADATA_FIELD_NUMBER: builtins.int
    SOURCE_FILE_FIELD_NUMBER: builtins.int
    STACK_FRAME_WITH_ID_FIELD_NUMBER: builtins.int
    GRAPH_OP_CREATION_FIELD_NUMBER: builtins.int
    DEBUGGED_GRAPH_FIELD_NUMBER: builtins.int
    EXECUTION_FIELD_NUMBER: builtins.int
    GRAPH_EXECUTION_TRACE_FIELD_NUMBER: builtins.int
    GRAPH_ID_FIELD_NUMBER: builtins.int
    DEBUGGED_DEVICE_FIELD_NUMBER: builtins.int
    wall_time: builtins.float
    """Timestamp in seconds (with microsecond precision)."""
    step: builtins.int
    """Step of training (if available)."""
    graph_id: builtins.str
    """The ID of the graph (i.e., FuncGraph) executed here: applicable only
    to the execution of a FuncGraph.
    """
    @property
    def debug_metadata(self) -> global___DebugMetadata:
        """Metadata related to this debugging data."""

    @property
    def source_file(self) -> global___SourceFile:
        """The content of a source file."""

    @property
    def stack_frame_with_id(self) -> global___StackFrameWithId:
        """A stack frame (filename, line number and column number, function name and
        code string) with ID.
        """

    @property
    def graph_op_creation(self) -> global___GraphOpCreation:
        """The creation of an op within a graph (e.g., a FuncGraph compiled from
        a Python function).
        """

    @property
    def debugged_graph(self) -> global___DebuggedGraph:
        """Information about a debugged graph."""

    @property
    def execution(self) -> global___Execution:
        """Execution of an op or a Graph (e.g., a tf.function)."""

    @property
    def graph_execution_trace(self) -> global___GraphExecutionTrace:
        """A graph execution trace: Contains information about the intermediate
        tensors computed during the graph execution.
        """

    @property
    def debugged_device(self) -> global___DebuggedDevice:
        """A device on which debugger-instrumented ops and/or tensors reside."""

    def __init__(
        self,
        *,
        wall_time: builtins.float | None = ...,
        step: builtins.int | None = ...,
        debug_metadata: global___DebugMetadata | None = ...,
        source_file: global___SourceFile | None = ...,
        stack_frame_with_id: global___StackFrameWithId | None = ...,
        graph_op_creation: global___GraphOpCreation | None = ...,
        debugged_graph: global___DebuggedGraph | None = ...,
        execution: global___Execution | None = ...,
        graph_execution_trace: global___GraphExecutionTrace | None = ...,
        graph_id: builtins.str | None = ...,
        debugged_device: global___DebuggedDevice | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing.Literal[
            "debug_metadata",
            b"debug_metadata",
            "debugged_device",
            b"debugged_device",
            "debugged_graph",
            b"debugged_graph",
            "execution",
            b"execution",
            "graph_execution_trace",
            b"graph_execution_trace",
            "graph_id",
            b"graph_id",
            "graph_op_creation",
            b"graph_op_creation",
            "source_file",
            b"source_file",
            "stack_frame_with_id",
            b"stack_frame_with_id",
            "what",
            b"what",
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "debug_metadata",
            b"debug_metadata",
            "debugged_device",
            b"debugged_device",
            "debugged_graph",
            b"debugged_graph",
            "execution",
            b"execution",
            "graph_execution_trace",
            b"graph_execution_trace",
            "graph_id",
            b"graph_id",
            "graph_op_creation",
            b"graph_op_creation",
            "source_file",
            b"source_file",
            "stack_frame_with_id",
            b"stack_frame_with_id",
            "step",
            b"step",
            "wall_time",
            b"wall_time",
            "what",
            b"what",
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing.Literal["what", b"what"]
    ) -> (
        typing.Literal[
            "debug_metadata",
            "source_file",
            "stack_frame_with_id",
            "graph_op_creation",
            "debugged_graph",
            "execution",
            "graph_execution_trace",
            "graph_id",
            "debugged_device",
        ]
        | None
    ): ...

global___DebugEvent = DebugEvent

@typing.final
class DebugMetadata(google.protobuf.message.Message):
    """Metadata about the debugger and the debugged TensorFlow program."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TENSORFLOW_VERSION_FIELD_NUMBER: builtins.int
    FILE_VERSION_FIELD_NUMBER: builtins.int
    TFDBG_RUN_ID_FIELD_NUMBER: builtins.int
    tensorflow_version: builtins.str
    """Version of TensorFlow."""
    file_version: builtins.str
    """Version of the DebugEvent file format.
    Has a format of "debug.Event:<number>", e.g., "debug.Event:1".
    """
    tfdbg_run_id: builtins.str
    """A unique ID for the current run of tfdbg.
    A run of tfdbg is defined as a TensorFlow job instrumented by tfdbg.
    Multiple hosts in a distributed TensorFlow job instrumented by tfdbg
    have the same ID.
    """
    def __init__(
        self,
        *,
        tensorflow_version: builtins.str | None = ...,
        file_version: builtins.str | None = ...,
        tfdbg_run_id: builtins.str | None = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "file_version", b"file_version", "tensorflow_version", b"tensorflow_version", "tfdbg_run_id", b"tfdbg_run_id"
        ],
    ) -> None: ...

global___DebugMetadata = DebugMetadata

@typing.final
class SourceFile(google.protobuf.message.Message):
    """Content of a source file involved in the execution of the debugged TensorFlow
    program.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FILE_PATH_FIELD_NUMBER: builtins.int
    HOST_NAME_FIELD_NUMBER: builtins.int
    LINES_FIELD_NUMBER: builtins.int
    file_path: builtins.str
    """Path to the file."""
    host_name: builtins.str
    """Name of the host on which the file is located."""
    @property
    def lines(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Line-by-line content of the file."""

    def __init__(
        self,
        *,
        file_path: builtins.str | None = ...,
        host_name: builtins.str | None = ...,
        lines: collections.abc.Iterable[builtins.str] | None = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing.Literal["file_path", b"file_path", "host_name", b"host_name", "lines", b"lines"]
    ) -> None: ...

global___SourceFile = SourceFile

@typing.final
class StackFrameWithId(google.protobuf.message.Message):
    """A stack frame with ID."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    ID_FIELD_NUMBER: builtins.int
    FILE_LINE_COL_FIELD_NUMBER: builtins.int
    id: builtins.str
    """A unique ID for the stack frame: A UUID-like string."""
    @property
    def file_line_col(self) -> tensorflow.core.framework.graph_debug_info_pb2.GraphDebugInfo.FileLineCol:
        """Stack frame, i.e., a frame of a stack trace, containing information
        regarding the file name, line number, function name, code content
        of the line, and column number (if available).
        """

    def __init__(
        self,
        *,
        id: builtins.str | None = ...,
        file_line_col: tensorflow.core.framework.graph_debug_info_pb2.GraphDebugInfo.FileLineCol | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["file_line_col", b"file_line_col"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing.Literal["file_line_col", b"file_line_col", "id", b"id"]) -> None: ...

global___StackFrameWithId = StackFrameWithId

@typing.final
class CodeLocation(google.protobuf.message.Message):
    """Code location information: A stack trace with host-name information.
    Instead of encoding the detailed stack trace, this proto refers to IDs of
    stack frames stored as `StackFrameWithId` protos.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    HOST_NAME_FIELD_NUMBER: builtins.int
    STACK_FRAME_IDS_FIELD_NUMBER: builtins.int
    host_name: builtins.str
    """Host name on which the source files are located."""
    @property
    def stack_frame_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """ID to a stack frame, each of which is pointed to
        by a unique ID. The ordering of the frames is consistent with Python's
        `traceback.extract_tb()`.
        """

    def __init__(
        self, *, host_name: builtins.str | None = ..., stack_frame_ids: collections.abc.Iterable[builtins.str] | None = ...
    ) -> None: ...
    def ClearField(
        self, field_name: typing.Literal["host_name", b"host_name", "stack_frame_ids", b"stack_frame_ids"]
    ) -> None: ...

global___CodeLocation = CodeLocation

@typing.final
class GraphOpCreation(google.protobuf.message.Message):
    """The creation of an op in a TensorFlow Graph (e.g., FuncGraph in TF2)."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    OP_TYPE_FIELD_NUMBER: builtins.int
    OP_NAME_FIELD_NUMBER: builtins.int
    GRAPH_NAME_FIELD_NUMBER: builtins.int
    GRAPH_ID_FIELD_NUMBER: builtins.int
    DEVICE_NAME_FIELD_NUMBER: builtins.int
    INPUT_NAMES_FIELD_NUMBER: builtins.int
    NUM_OUTPUTS_FIELD_NUMBER: builtins.int
    CODE_LOCATION_FIELD_NUMBER: builtins.int
    OUTPUT_TENSOR_IDS_FIELD_NUMBER: builtins.int
    op_type: builtins.str
    """Type of the op (e.g., "MatMul")."""
    op_name: builtins.str
    """Name of the op (e.g., "Dense/MatMul_1")."""
    graph_name: builtins.str
    """Name of the graph that the op is a part of (if available)."""
    graph_id: builtins.str
    """Unique ID of the graph (generated by debugger).
    This is the ID of the immediately-enclosing graph.
    """
    device_name: builtins.str
    """Name of the device that the op is assigned to (if available)."""
    num_outputs: builtins.int
    """Number of output tensors emitted by the op."""
    @property
    def input_names(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Names of the input tensors to the op."""

    @property
    def code_location(self) -> global___CodeLocation:
        """The unique ID for code location (stack trace) of the op's creation."""

    @property
    def output_tensor_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """Unique IDs for the output tensors of this op."""

    def __init__(
        self,
        *,
        op_type: builtins.str | None = ...,
        op_name: builtins.str | None = ...,
        graph_name: builtins.str | None = ...,
        graph_id: builtins.str | None = ...,
        device_name: builtins.str | None = ...,
        input_names: collections.abc.Iterable[builtins.str] | None = ...,
        num_outputs: builtins.int | None = ...,
        code_location: global___CodeLocation | None = ...,
        output_tensor_ids: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["code_location", b"code_location"]) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "code_location",
            b"code_location",
            "device_name",
            b"device_name",
            "graph_id",
            b"graph_id",
            "graph_name",
            b"graph_name",
            "input_names",
            b"input_names",
            "num_outputs",
            b"num_outputs",
            "op_name",
            b"op_name",
            "op_type",
            b"op_type",
            "output_tensor_ids",
            b"output_tensor_ids",
        ],
    ) -> None: ...

global___GraphOpCreation = GraphOpCreation

@typing.final
class DebuggedGraph(google.protobuf.message.Message):
    """A debugger-instrumented graph."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    GRAPH_ID_FIELD_NUMBER: builtins.int
    GRAPH_NAME_FIELD_NUMBER: builtins.int
    INSTRUMENTED_OPS_FIELD_NUMBER: builtins.int
    ORIGINAL_GRAPH_DEF_FIELD_NUMBER: builtins.int
    INSTRUMENTED_GRAPH_DEF_FIELD_NUMBER: builtins.int
    OUTER_CONTEXT_ID_FIELD_NUMBER: builtins.int
    graph_id: builtins.str
    """An ID for the graph.
    This can be used up to look up graph names. Generated by the debugger.
    """
    graph_name: builtins.str
    """Name of the graph (if available)."""
    original_graph_def: builtins.bytes
    """Original (uninstrumented) GraphDef (if available)."""
    instrumented_graph_def: builtins.bytes
    """An encoded version of a GraphDef.
    This graph may include the debugger-inserted ops.
    """
    outer_context_id: builtins.str
    """IDs of the immediate enclosing context (graph), if any."""
    @property
    def instrumented_ops(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """Names of the instrumented ops. This can be used to look up op name
        based on the numeric-summary tensors (2nd column).
        """

    def __init__(
        self,
        *,
        graph_id: builtins.str | None = ...,
        graph_name: builtins.str | None = ...,
        instrumented_ops: collections.abc.Iterable[builtins.str] | None = ...,
        original_graph_def: builtins.bytes | None = ...,
        instrumented_graph_def: builtins.bytes | None = ...,
        outer_context_id: builtins.str | None = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "graph_id",
            b"graph_id",
            "graph_name",
            b"graph_name",
            "instrumented_graph_def",
            b"instrumented_graph_def",
            "instrumented_ops",
            b"instrumented_ops",
            "original_graph_def",
            b"original_graph_def",
            "outer_context_id",
            b"outer_context_id",
        ],
    ) -> None: ...

global___DebuggedGraph = DebuggedGraph

@typing.final
class DebuggedDevice(google.protobuf.message.Message):
    """A device on which ops and/or tensors are instrumented by the debugger."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DEVICE_NAME_FIELD_NUMBER: builtins.int
    DEVICE_ID_FIELD_NUMBER: builtins.int
    device_name: builtins.str
    """Name of the device."""
    device_id: builtins.int
    """A debugger-generated ID for the device. Guaranteed to be unique within
    the scope of the debugged TensorFlow program, including single-host and
    multi-host settings.
    TODO(cais): Test the uniqueness guarantee in multi-host settings.
    """
    def __init__(self, *, device_name: builtins.str | None = ..., device_id: builtins.int | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["device_id", b"device_id", "device_name", b"device_name"]) -> None: ...

global___DebuggedDevice = DebuggedDevice

@typing.final
class Execution(google.protobuf.message.Message):
    """Data relating to the eager execution of an op or a Graph.
    For a op that generates N output tensors (N >= 0), only one
    Execution proto will be used to describe the execution event.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    OP_TYPE_FIELD_NUMBER: builtins.int
    NUM_OUTPUTS_FIELD_NUMBER: builtins.int
    GRAPH_ID_FIELD_NUMBER: builtins.int
    INPUT_TENSOR_IDS_FIELD_NUMBER: builtins.int
    OUTPUT_TENSOR_IDS_FIELD_NUMBER: builtins.int
    TENSOR_DEBUG_MODE_FIELD_NUMBER: builtins.int
    TENSOR_PROTOS_FIELD_NUMBER: builtins.int
    CODE_LOCATION_FIELD_NUMBER: builtins.int
    OUTPUT_TENSOR_DEVICE_IDS_FIELD_NUMBER: builtins.int
    op_type: builtins.str
    """Op type (e.g., "MatMul").
    In the case of a Graph, this is the name of the Graph.
    """
    num_outputs: builtins.int
    """Number of output tensors."""
    graph_id: builtins.str
    """The graph that's executed: applicable only to the eager
    execution of a FuncGraph.
    """
    tensor_debug_mode: global___TensorDebugMode.ValueType
    """Type of the tensor value encapsulated in this proto."""
    @property
    def input_tensor_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """IDs of the input tensors (if available)."""

    @property
    def output_tensor_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """IDs of the output tensors (if availbable).
        If specified, must have the same length as tensor_protos.
        """

    @property
    def tensor_protos(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[tensorflow.core.framework.tensor_pb2.TensorProto]:
        """Output Tensor values in the type described by `tensor_value_type`.
        The length of this should match `num_outputs`.
        """

    @property
    def code_location(self) -> global___CodeLocation:
        """Stack trace of the eager execution."""

    @property
    def output_tensor_device_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """Debugged-generated IDs of the devices on which the output tensors reside.
        To look up details about the device (e.g., name), cross-reference this
        field with the DebuggedDevice messages.
        """

    def __init__(
        self,
        *,
        op_type: builtins.str | None = ...,
        num_outputs: builtins.int | None = ...,
        graph_id: builtins.str | None = ...,
        input_tensor_ids: collections.abc.Iterable[builtins.int] | None = ...,
        output_tensor_ids: collections.abc.Iterable[builtins.int] | None = ...,
        tensor_debug_mode: global___TensorDebugMode.ValueType | None = ...,
        tensor_protos: collections.abc.Iterable[tensorflow.core.framework.tensor_pb2.TensorProto] | None = ...,
        code_location: global___CodeLocation | None = ...,
        output_tensor_device_ids: collections.abc.Iterable[builtins.int] | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["code_location", b"code_location"]) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "code_location",
            b"code_location",
            "graph_id",
            b"graph_id",
            "input_tensor_ids",
            b"input_tensor_ids",
            "num_outputs",
            b"num_outputs",
            "op_type",
            b"op_type",
            "output_tensor_device_ids",
            b"output_tensor_device_ids",
            "output_tensor_ids",
            b"output_tensor_ids",
            "tensor_debug_mode",
            b"tensor_debug_mode",
            "tensor_protos",
            b"tensor_protos",
        ],
    ) -> None: ...

global___Execution = Execution

@typing.final
class GraphExecutionTrace(google.protobuf.message.Message):
    """Data relating to an execution of a Graph (e.g., an eager execution of a
    FuncGraph).
    The values of the intermediate tensors computed in the graph are recorded
    in this proto. A graph execution may correspond to one or more pieces of
    `GraphExecutionTrace`, depending on whether the instrumented tensor values
    are summarized in an aggregated or separate fashion.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TFDBG_CONTEXT_ID_FIELD_NUMBER: builtins.int
    OP_NAME_FIELD_NUMBER: builtins.int
    OUTPUT_SLOT_FIELD_NUMBER: builtins.int
    TENSOR_DEBUG_MODE_FIELD_NUMBER: builtins.int
    TENSOR_PROTO_FIELD_NUMBER: builtins.int
    DEVICE_NAME_FIELD_NUMBER: builtins.int
    tfdbg_context_id: builtins.str
    """Unique ID of the context that the executed op(s) belong to (e.g., a
    compiled concrete tf.function).
    """
    op_name: builtins.str
    """Name of the op (applicable only in the case of the `FULL_TENSOR` trace
    level).
    """
    output_slot: builtins.int
    """Output slot of the tensor (applicable only in the case of the `FULL_TENSOR`
    trace level).
    """
    tensor_debug_mode: global___TensorDebugMode.ValueType
    """Type of the tensor value encapsulated in this proto."""
    device_name: builtins.str
    """Name of the device that the op belongs to."""
    @property
    def tensor_proto(self) -> tensorflow.core.framework.tensor_pb2.TensorProto:
        """Tensor value in the type described by `tensor_value_type`.
        This tensor may summarize the value of a single intermediate op of the
        graph, or those of multiple intermediate tensors.
        """

    def __init__(
        self,
        *,
        tfdbg_context_id: builtins.str | None = ...,
        op_name: builtins.str | None = ...,
        output_slot: builtins.int | None = ...,
        tensor_debug_mode: global___TensorDebugMode.ValueType | None = ...,
        tensor_proto: tensorflow.core.framework.tensor_pb2.TensorProto | None = ...,
        device_name: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["tensor_proto", b"tensor_proto"]) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "device_name",
            b"device_name",
            "op_name",
            b"op_name",
            "output_slot",
            b"output_slot",
            "tensor_debug_mode",
            b"tensor_debug_mode",
            "tensor_proto",
            b"tensor_proto",
            "tfdbg_context_id",
            b"tfdbg_context_id",
        ],
    ) -> None: ...

global___GraphExecutionTrace = GraphExecutionTrace



from .claims import (
    CodeIDToken as CodeIDToken,
    HybridIDToken as HybridIDToken,
    IDToken as IDToken,
    ImplicitIDToken as ImplicitIDToken,
    UserInfo as UserInfo,
    get_claim_cls_by_response_type as get_claim_cls_by_response_type,
)
from .grants import (
    OpenIDCode as OpenIDCode,
    OpenIDHybridGrant as OpenIDHybridGrant,
    OpenIDImplicitGrant as OpenIDImplicitGrant,
    OpenIDToken as OpenIDToken,
)
from .models import AuthorizationCodeMixin as AuthorizationCodeMixin
from .userinfo import UserInfoEndpoint as UserInfoEndpoint

__all__ = [
    "AuthorizationCodeMixin",
    "IDToken",
    "CodeIDToken",
    "ImplicitIDToken",
    "HybridIDToken",
    "UserInfo",
    "UserInfoEndpoint",
    "get_claim_cls_by_response_type",
    "OpenIDToken",
    "OpenIDCode",
    "OpenIDHybridGrant",
    "OpenIDImplicitGrant",
]



import re
from typing import ClassVar, TypeVar

import click

_T = TypeVar("_T")

class EmailParamType(click.ParamType):
    EMAIL_REGEX: ClassVar[re.Pattern[str]]
    def convert(self, value: str, param: click.Parameter | None, ctx: click.Context | None) -> str: ...

class PasswordParamType(click.ParamType):
    def convert(self, value: _T, param: click.Parameter | None, ctx: click.Context | None) -> _T: ...

class TextAreaParamType(click.ParamType):
    def convert(self, value: _T, param: click.Parameter | None, ctx: click.Context | None) -> _T: ...

EMAIL_TYPE: EmailParamType
PASSWORD_TYPE: PasswordParamType
TEXTAREA_TYPE: TextAreaParamType



from collections.abc import Hashable
from typing import TypeVar

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

_G = TypeVar("_G", bound=Graph[Hashable])

__all__ = ["complement", "reverse"]

@_dispatchable
def complement(G: Graph[_Node]): ...
@_dispatchable
def reverse(G: _G, copy: bool = True) -> _G: ...



from _typeshed import Incomplete

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = [
    "geometric_edges",
    "geographical_threshold_graph",
    "navigable_small_world_graph",
    "random_geometric_graph",
    "soft_random_geometric_graph",
    "thresholded_random_geometric_graph",
    "waxman_graph",
    "geometric_soft_configuration_graph",
]

@_dispatchable
def geometric_edges(G: Graph[_Node], radius, p: float = 2): ...
@_dispatchable
def random_geometric_graph(n, radius, dim: int = 2, pos=None, p: float = 2, seed=None): ...
@_dispatchable
def soft_random_geometric_graph(n, radius, dim: int = 2, pos=None, p: float = 2, p_dist=None, seed=None): ...
@_dispatchable
def geographical_threshold_graph(n, theta, dim: int = 2, pos=None, weight=None, metric=None, p_dist=None, seed=None): ...
@_dispatchable
def waxman_graph(
    n, beta: float = 0.4, alpha: float = 0.1, L=None, domain=(0, 0, 1, 1), metric=None, seed=None
) -> Graph[Incomplete]: ...

# docstring marks p as int, but it still works with floats. So I think it's better for consistency
@_dispatchable
def navigable_small_world_graph(n, p: float = 1, q: int = 1, r: float = 2, dim: int = 2, seed=None): ...
@_dispatchable
def thresholded_random_geometric_graph(n, radius, theta, dim: int = 2, pos=None, weight=None, p: float = 2, seed=None): ...
@_dispatchable
def geometric_soft_configuration_graph(
    *, beta, n=None, gamma=None, mean_degree=None, kappas=None, seed=None
) -> Graph[Incomplete]: ...



from logging import Logger
from typing import Literal

from oauthlib.common import Request, _HTTPMethod

from ..request_validator import RequestValidator
from .base import BaseEndpoint

log: Logger

class RevocationEndpoint(BaseEndpoint):
    valid_token_types: tuple[Literal["access_token"], Literal["refresh_token"]]
    valid_request_methods: tuple[Literal["POST"]]
    request_validator: RequestValidator
    supported_token_types: tuple[str, ...]
    enable_jsonp: bool
    def __init__(
        self,
        request_validator: RequestValidator,
        supported_token_types: tuple[str, ...] | None = None,
        enable_jsonp: bool = False,
    ) -> None: ...
    def create_revocation_response(
        self, uri: str, http_method: _HTTPMethod = "POST", body: str | None = None, headers: dict[str, str] | None = None
    ) -> tuple[dict[str, str], str, int]: ...
    def validate_revocation_request(self, request: Request) -> None: ...



from typing import Any

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["adjacency_data", "adjacency_graph"]

# Any: Complex type union
def adjacency_data(G: Graph[_Node], attrs={"id": "id", "key": "key"}) -> dict[str, Any]: ...
@_dispatchable
def adjacency_graph(data, directed: bool = False, multigraph: bool = True, attrs={"id": "id", "key": "key"}): ...



import types
import xml.etree.ElementTree as ET
from _typeshed import Incomplete
from collections.abc import Generator
from contextlib import contextmanager
from typing import Any

class LxmlSyntaxError(Exception): ...

class _IncrementalFileWriter:
    global_nsmap: dict[str, str]
    is_html: bool
    def __init__(self, output_file: ET._FileWrite) -> None: ...
    @contextmanager
    def element(
        self,
        tag: str | ET._ElementCallable,
        attrib: dict[str, str] | None = None,
        nsmap: dict[str, str] | None = None,
        **_extra: str,
    ) -> Generator[None]: ...
    def write(self, arg: str | ET.Element[Any]) -> None: ...
    def __enter__(self) -> None: ...
    def __exit__(
        self, type: type[BaseException] | None, value: BaseException | None, traceback: types.TracebackType | None
    ) -> None: ...

class xmlfile:
    encoding: str
    writer_cm: Incomplete
    def __init__(
        self, output_file: ET._FileWrite, buffered: bool = False, encoding: str = "utf-8", close: bool = False
    ) -> None: ...
    def __enter__(self) -> _IncrementalFileWriter: ...
    def __exit__(
        self, type: type[BaseException] | None, value: BaseException | None, traceback: types.TracebackType | None
    ) -> None: ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import tensorflow.core.framework.resource_handle_pb2
import tensorflow.core.framework.tensor_shape_pb2
import tensorflow.core.framework.types_pb2

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class TensorProto(google.protobuf.message.Message):
    """Protocol buffer representing a tensor."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    DTYPE_FIELD_NUMBER: builtins.int
    TENSOR_SHAPE_FIELD_NUMBER: builtins.int
    VERSION_NUMBER_FIELD_NUMBER: builtins.int
    TENSOR_CONTENT_FIELD_NUMBER: builtins.int
    HALF_VAL_FIELD_NUMBER: builtins.int
    FLOAT_VAL_FIELD_NUMBER: builtins.int
    DOUBLE_VAL_FIELD_NUMBER: builtins.int
    INT_VAL_FIELD_NUMBER: builtins.int
    STRING_VAL_FIELD_NUMBER: builtins.int
    SCOMPLEX_VAL_FIELD_NUMBER: builtins.int
    INT64_VAL_FIELD_NUMBER: builtins.int
    BOOL_VAL_FIELD_NUMBER: builtins.int
    DCOMPLEX_VAL_FIELD_NUMBER: builtins.int
    RESOURCE_HANDLE_VAL_FIELD_NUMBER: builtins.int
    VARIANT_VAL_FIELD_NUMBER: builtins.int
    UINT32_VAL_FIELD_NUMBER: builtins.int
    UINT64_VAL_FIELD_NUMBER: builtins.int
    FLOAT8_VAL_FIELD_NUMBER: builtins.int
    dtype: tensorflow.core.framework.types_pb2.DataType.ValueType
    """Data type of the tensor."""
    version_number: builtins.int
    """Only one of the representations below is set, one of "tensor_contents" and
    the "xxx_val" attributes.  We are not using oneof because as oneofs cannot
    contain repeated fields it would require another extra set of messages.

    Version number.

    In version 0, if the "repeated xxx" representations contain only one
    element, that element is repeated to fill the shape.  This makes it easy
    to represent a constant Tensor with a single value.
    """
    tensor_content: builtins.bytes
    """Serialized raw tensor content from either Tensor::AsProtoTensorContent or
    memcpy in tensorflow::grpc::EncodeTensorToByteBuffer. This representation
    can be used for all tensor types. The purpose of this representation is to
    reduce serialization overhead during RPC call by avoiding serialization of
    many repeated small items.
    """
    float8_val: builtins.bytes
    """DT_FLOAT8_*, use variable-sized set of bytes
    (i.e. the equivalent of repeated uint8, if such a thing existed).
    """
    @property
    def tensor_shape(self) -> tensorflow.core.framework.tensor_shape_pb2.TensorShapeProto:
        """Shape of the tensor.  TODO(touts): sort out the 0-rank issues."""

    @property
    def half_val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """Type specific representations that make it easy to create tensor protos in
        all languages.  Only the representation corresponding to "dtype" can
        be set.  The values hold the flattened representation of the tensor in
        row major order.

        DT_HALF, DT_BFLOAT16. Note that since protobuf has no int16 type, we'll
        have some pointless zero padding for each value here.
        """

    @property
    def float_val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]:
        """DT_FLOAT."""

    @property
    def double_val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]:
        """DT_DOUBLE."""

    @property
    def int_val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """DT_INT32, DT_INT16, DT_UINT16, DT_INT8, DT_UINT8."""

    @property
    def string_val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]:
        """DT_STRING"""

    @property
    def scomplex_val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]:
        """DT_COMPLEX64. scomplex_val(2*i) and scomplex_val(2*i+1) are real
        and imaginary parts of i-th single precision complex.
        """

    @property
    def int64_val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """DT_INT64"""

    @property
    def bool_val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bool]:
        """DT_BOOL"""

    @property
    def dcomplex_val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]:
        """DT_COMPLEX128. dcomplex_val(2*i) and dcomplex_val(2*i+1) are real
        and imaginary parts of i-th double precision complex.
        """

    @property
    def resource_handle_val(
        self,
    ) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[
        tensorflow.core.framework.resource_handle_pb2.ResourceHandleProto
    ]:
        """DT_RESOURCE"""

    @property
    def variant_val(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___VariantTensorDataProto]:
        """DT_VARIANT"""

    @property
    def uint32_val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """DT_UINT32"""

    @property
    def uint64_val(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """DT_UINT64"""

    def __init__(
        self,
        *,
        dtype: tensorflow.core.framework.types_pb2.DataType.ValueType | None = ...,
        tensor_shape: tensorflow.core.framework.tensor_shape_pb2.TensorShapeProto | None = ...,
        version_number: builtins.int | None = ...,
        tensor_content: builtins.bytes | None = ...,
        half_val: collections.abc.Iterable[builtins.int] | None = ...,
        float_val: collections.abc.Iterable[builtins.float] | None = ...,
        double_val: collections.abc.Iterable[builtins.float] | None = ...,
        int_val: collections.abc.Iterable[builtins.int] | None = ...,
        string_val: collections.abc.Iterable[builtins.bytes] | None = ...,
        scomplex_val: collections.abc.Iterable[builtins.float] | None = ...,
        int64_val: collections.abc.Iterable[builtins.int] | None = ...,
        bool_val: collections.abc.Iterable[builtins.bool] | None = ...,
        dcomplex_val: collections.abc.Iterable[builtins.float] | None = ...,
        resource_handle_val: (
            collections.abc.Iterable[tensorflow.core.framework.resource_handle_pb2.ResourceHandleProto] | None
        ) = ...,
        variant_val: collections.abc.Iterable[global___VariantTensorDataProto] | None = ...,
        uint32_val: collections.abc.Iterable[builtins.int] | None = ...,
        uint64_val: collections.abc.Iterable[builtins.int] | None = ...,
        float8_val: builtins.bytes | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["tensor_shape", b"tensor_shape"]) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "bool_val",
            b"bool_val",
            "dcomplex_val",
            b"dcomplex_val",
            "double_val",
            b"double_val",
            "dtype",
            b"dtype",
            "float8_val",
            b"float8_val",
            "float_val",
            b"float_val",
            "half_val",
            b"half_val",
            "int64_val",
            b"int64_val",
            "int_val",
            b"int_val",
            "resource_handle_val",
            b"resource_handle_val",
            "scomplex_val",
            b"scomplex_val",
            "string_val",
            b"string_val",
            "tensor_content",
            b"tensor_content",
            "tensor_shape",
            b"tensor_shape",
            "uint32_val",
            b"uint32_val",
            "uint64_val",
            b"uint64_val",
            "variant_val",
            b"variant_val",
            "version_number",
            b"version_number",
        ],
    ) -> None: ...

global___TensorProto = TensorProto

@typing.final
class VariantTensorDataProto(google.protobuf.message.Message):
    """Protocol buffer representing the serialization format of DT_VARIANT tensors."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TYPE_NAME_FIELD_NUMBER: builtins.int
    METADATA_FIELD_NUMBER: builtins.int
    TENSORS_FIELD_NUMBER: builtins.int
    type_name: builtins.str
    """Name of the type of objects being serialized."""
    metadata: builtins.bytes
    """Portions of the object that are not Tensors."""
    @property
    def tensors(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___TensorProto]:
        """Tensors contained within objects being serialized."""

    def __init__(
        self,
        *,
        type_name: builtins.str | None = ...,
        metadata: builtins.bytes | None = ...,
        tensors: collections.abc.Iterable[global___TensorProto] | None = ...,
    ) -> None: ...
    def ClearField(
        self, field_name: typing.Literal["metadata", b"metadata", "tensors", b"tensors", "type_name", b"type_name"]
    ) -> None: ...

global___VariantTensorDataProto = VariantTensorDataProto
